import os

import numpy as np
import pytest

from autolens import exc
from autolens.imaging import imaging_util as util

test_data_dir = "{}/../test_files/array/".format(os.path.dirname(os.path.realpath(__file__)))


@pytest.fixture(name="memoizer")
def make_memoizer():
    return util.Memoizer()


class TestMemoizer(object):
    def test_storing(self, memoizer):
        @memoizer
        def func(arg):
            return "result for {}".format(arg)

        func(1)
        func(2)
        func(1)

        assert memoizer.results == {"('arg', 1)": "result for 1", "('arg', 2)": "result for 2"}
        assert memoizer.calls == 2

    def test_multiple_arguments(self, memoizer):
        @memoizer
        def func(arg1, arg2):
            return arg1 * arg2

        func(1, 2)
        func(2, 1)
        func(1, 2)

        assert memoizer.results == {"('arg1', 1), ('arg2', 2)": 2, "('arg1', 2), ('arg2', 1)": 2}
        assert memoizer.calls == 2

    def test_key_word_arguments(self, memoizer):
        @memoizer
        def func(arg1=0, arg2=0):
            return arg1 * arg2

        func(arg1=1)
        func(arg2=1)
        func(arg1=1)
        func(arg1=1, arg2=1)

        assert memoizer.results == {"('arg1', 1)": 0, "('arg2', 1)": 0, "('arg1', 1), ('arg2', 1)": 1}
        assert memoizer.calls == 3

    def test_key_word_for_positional(self, memoizer):
        @memoizer
        def func(arg):
            return "result for {}".format(arg)

        func(1)
        func(arg=2)
        func(arg=1)

        assert memoizer.calls == 2

    def test_methods(self, memoizer):
        class Class(object):
            def __init__(self, value):
                self.value = value

            @memoizer
            def method(self):
                return self.value

        one = Class(1)
        two = Class(2)

        assert one.method() == 1
        assert two.method() == 2


class TestTotalPixels:

    def test__total_image_pixels_from_mask(self):
        mask = np.array([[True, False, True],
                         [False, False, False],
                         [True, False, True]])

        assert util.total_image_pixels_from_mask(mask) == 5

    def test__total_sub_pixels_from_mask(self):
        mask = np.array([[True, False, True],
                         [False, False, False],
                         [True, False, True]])

        assert util.total_sub_pixels_from_mask_and_sub_grid_size(mask, sub_grid_size=2) == 20

    def test__total_border_pixels_from_mask(self):
        mask = np.array([[True, True, True, True, True],
                         [True, False, False, False, True],
                         [True, False, False, False, True],
                         [True, False, False, False, True],
                         [True, True, True, True, True]])

        assert util.total_border_pixels_from_mask(mask) == 8


class TestGrid2d:

    def test__array_3x3__sets_up_arcsecond_grid(self):

        grid_2d = util.image_grid_2d_from_shape_and_pixel_scales(shape=(3, 3), pixel_scales=(2.0, 1.0))

        assert (grid_2d == np.array([[[2., -1.], [2., 0.], [2., 1.]],
                                     [[0., -1.], [0., 0.], [0., 1.]],
                                     [[-2., -1.], [-2., 0.], [-2., 1.]]])).all()

    def test__array_4x4_and_different_pixel_scale__sets_up_arcsecond_grid(self):

        grid_2d = util.image_grid_2d_from_shape_and_pixel_scales(shape=(4, 4), pixel_scales=(0.5, 0.5))

        assert (grid_2d == np.array([[[0.75, -0.75], [0.75, -0.25], [0.75, 0.25], [0.75, 0.75]],
                                     [[0.25, -0.75], [0.25, -0.25], [0.25, 0.25], [0.25, 0.75]],
                                     [[-0.25, -0.75], [-0.25, -0.25], [-0.25, 0.25], [-0.25, 0.75]],
                                     [[-0.75, -0.75], [-0.75, -0.25], [-0.75, 0.25], [-0.75, 0.75]]])).all()

    def test__array_2x3__sets_up_arcsecond_grid(self):
        grid_2d = util.image_grid_2d_from_shape_and_pixel_scales(shape=(2, 3), pixel_scales=(1.0, 1.0))

        assert (grid_2d == np.array([[[0.5, -1.], [0.5, 0.], [0.5, 1.]],
                                     [[-0.5, -1.], [-0.5, 0.], [-0.5, 1.]]])).all()

    def test__array_3x2__sets_up_arcsecond_grid(self):
        grid_2d = util.image_grid_2d_from_shape_and_pixel_scales(shape=(3, 2), pixel_scales=(1.0, 1.0))

        assert (grid_2d == np.array([[[1., -0.5], [1., 0.5]],
                                     [[0., -0.5], [0., 0.5]],
                                     [[-1., -0.5], [-1., 0.5]]])).all()


class TestGrid1d:

    def test__array_3x3__sets_up_arcsecond_grid(self):

        grid_2d = util.image_grid_1d_from_shape_and_pixel_scales(shape=(3, 3), pixel_scales=(2.0, 1.0))

        assert (grid_2d == np.array([[2., -1.], [2., 0.], [2., 1.],
                                     [0., -1.], [0., 0.], [0., 1.],
                                     [-2., -1.], [-2., 0.], [-2., 1.]])).all()

    def test__array_4x4_and_different_pixel_scale__sets_up_arcsecond_grid(self):

        grid_2d = util.image_grid_1d_from_shape_and_pixel_scales(shape=(4, 4), pixel_scales=(0.5, 0.5))

        assert (grid_2d == np.array([[0.75, -0.75], [0.75, -0.25], [0.75, 0.25], [0.75, 0.75],
                                     [0.25, -0.75], [0.25, -0.25], [0.25, 0.25], [0.25, 0.75],
                                     [-0.25, -0.75], [-0.25, -0.25], [-0.25, 0.25], [-0.25, 0.75],
                                     [-0.75, -0.75], [-0.75, -0.25], [-0.75, 0.25], [-0.75, 0.75]])).all()

    def test__array_2x3__sets_up_arcsecond_grid(self):
        grid_2d = util.image_grid_1d_from_shape_and_pixel_scales(shape=(2, 3), pixel_scales=(1.0, 1.0))

        assert (grid_2d == np.array([[0.5, -1.], [0.5, 0.], [0.5, 1.],
                                     [-0.5, -1.], [-0.5, 0.], [-0.5, 1.]])).all()

    def test__array_3x2__sets_up_arcsecond_grid(self):
        grid_2d = util.image_grid_1d_from_shape_and_pixel_scales(shape=(3, 2), pixel_scales=(1.0, 1.0))

        assert (grid_2d == np.array([[1., -0.5], [1., 0.5],
                                     [0., -0.5], [0., 0.5],
                                     [-1., -0.5], [-1., 0.5]])).all()


class TestImageGridMasked(object):

    def test__setup_3x3_image_1_coordinate_in_mask(self):
        mask = np.array([[True, True, True],
                         [True, False, True],
                         [True, True, True]])

        image_grid = util.image_grid_1d_masked_from_mask_and_pixel_scales(mask=mask, pixel_scales=(3.0, 6.0))

        assert (image_grid[0] == np.array([0.0, 0.0])).all()

    def test__setup_3x3_image__five_coordinates_in_mask(self):
        mask = np.array([[True, False, True],
                         [False, False, False],
                         [True, False, True]])

        image_grid = util.image_grid_1d_masked_from_mask_and_pixel_scales(mask=mask, pixel_scales=(6.0, 3.0))

        assert (image_grid == np.array([           [6., 0.],
                                        [0., -3.], [0., 0.], [0., 3.],
                                                   [-6., 0.]])).all()

    def test__setup_4x4_image__ten_coordinates_in_grid__new_pixel_scale(self):
        mask = np.array([[True, False, False, True],
                         [False, False, False, True],
                         [True, False, False, True],
                         [False, False, False, True]])

        image_grid = util.image_grid_1d_masked_from_mask_and_pixel_scales(mask=mask, pixel_scales=(1.0, 1.0))

        assert (image_grid == np.array(             [[1.5, -0.5], [1.5, 0.5],
                                       [0.5, -1.5], [0.5, -0.5], [0.5, 0.5],
                                                     [-0.5, -0.5], [-0.5, 0.5],
                                        [-1.5, -1.5], [-1.5, -0.5], [-1.5, 0.5]])).all()

    def test__setup_3x4_image__six_grid(self):
        mask = np.array([[True, False, True, True],
                         [False, False, False, True],
                         [True, False, True, False]])

        image_grid = util.image_grid_1d_masked_from_mask_and_pixel_scales(mask=mask, pixel_scales=(3.0, 3.0))

        assert (image_grid == np.array([             [3., -1.5],
                                       [0., -4.5], [0., -1.5], [0., 1.5],
                                                    [-3., -1.5],           [-3., 4.5]])).all()


class TestSubGridMasked(object):

    def test__3x3_mask_with_one_pixel__2x2_sub_grid__grid(self):
        mask = np.array([[True, True, True],
                         [True, False, True],
                         [True, True, True]])

        sub_grid = util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(mask=mask, pixel_scales=(3.0, 6.0),
                                                                                    sub_grid_size=2)

        assert (sub_grid[0:4] == np.array([[0.5, -1.0], [0.5, 1.0],
                                          [-0.5, -1.0], [-0.5, 1.0]])).all()

    def test__3x3_mask_with_row_of_pixels__2x2_sub_grid__grid(self):
        mask = np.array([[True, True, True],
                         [False, False, False],
                         [True, True, True]])

        sub_grid = util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(mask=mask, pixel_scales=(3.0, 3.0),
                                                                                    sub_grid_size=2)

        assert (sub_grid[0:4] == np.array([[0.5, -3.5], [0.5, -2.5],
                                           [-0.5, -3.5], [-0.5, -2.5]])).all()

        assert (sub_grid[4:8] == np.array([[0.5, -0.5], [0.5, 0.5],
                                           [-0.5, -0.5], [-0.5, 0.5]])).all()

        assert (sub_grid[8:12] == np.array([[0.5,  2.5], [0.5, 3.5],
                                            [-0.5, 2.5], [-0.5, 3.5]])).all()

    def test__3x3_mask_with_row_and_column_of_pixels__2x2_sub_grid__grid(self):
        mask = np.array([[True, True, False],
                         [False, False, False],
                         [True, True, False]])

        sub_grid = util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(mask=mask, pixel_scales=(3.0, 3.0),
                                                                                    sub_grid_size=2)

        assert (sub_grid == np.array([[3.5, 2.5], [3.5, 3.5], [2.5, 2.5], [2.5, 3.5],
                                      [0.5, -3.5], [0.5, -2.5], [-0.5, -3.5], [-0.5, -2.5],
                                      [0.5, -0.5], [0.5, 0.5], [-0.5, -0.5], [-0.5, 0.5],
                                      [0.5, 2.5], [0.5, 3.5], [-0.5, 2.5], [-0.5, 3.5],
                                      [-2.5, 2.5], [-2.5, 3.5], [-3.5, 2.5], [-3.5, 3.5]])).all()

    def test__3x3_mask_with_row_and_column_of_pixels__2x2_sub_grid__different_pixel_scale(self):
        mask = np.array([[True, True, False],
                         [False, False, False],
                         [True, True, False]])

        sub_grid = util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(mask=mask, pixel_scales=(0.3, 0.3),
                                                                                    sub_grid_size=2)

        sub_grid = np.round(sub_grid, decimals=2)

        np.testing.assert_almost_equal(sub_grid,
                                       np.array([[0.35, 0.25], [0.35, 0.35], [0.25, 0.25], [0.25, 0.35],
                                                 [0.05, -0.35], [0.05, -0.25], [-0.05, -0.35], [-0.05, -0.25],
                                                 [0.05, -0.05], [0.05, 0.05], [-0.05, -0.05], [-0.05, 0.05],
                                                 [0.05, 0.25], [0.05, 0.35], [-0.05, 0.25], [-0.05, 0.35],
                                                 [-0.25, 0.25], [-0.25, 0.35], [-0.35, 0.25], [-0.35, 0.35]]))

    def test__3x3_mask_with_one_pixel__3x3_sub_grid__grid(self):
        mask = np.array([[True, True, True],
                         [True, False, True],
                         [True, True, True]])

        sub_grid = util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(mask=mask, pixel_scales=(3.0, 3.0),
                                                                                    sub_grid_size=3)

        assert (sub_grid == np.array([[[0.75, -0.75], [0.75, 0.], [0.75, 0.75], [0., -0.75], [0., 0.],
                                       [0., 0.75], [-0.75, -0.75], [-0.75, 0.], [-0.75, 0.75]]])).all()

    def test__3x3_mask_with_one_row__3x3_sub_grid__grid(self):
        mask = np.array([[True, True, False],
                         [True, False, True],
                         [True, True, False]])

        sub_grid = util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(mask=mask, pixel_scales=(2.0, 2.0),
                                                                                    sub_grid_size=3)

        assert (sub_grid == np.array([[2.5, 1.5], [2.5, 2.], [2.5, 2.5],
                                      [2., 1.5], [2., 2.], [2., 2.5],
                                      [1.5, 1.5], [1.5, 2.], [1.5, 2.5],
                                      [0.5, -0.5], [0.5, 0.], [0.5, 0.5],
                                      [0., -0.5], [0., 0.], [0., 0.5],
                                      [-0.5, -0.5], [-0.5, 0.], [-0.5, 0.5],
                                      [-1.5, 1.5], [-1.5, 2.], [-1.5, 2.5],
                                      [-2., 1.5], [-2., 2.], [-2., 2.5],
                                      [-2.5, 1.5], [-2.5, 2.], [-2.5, 2.5]])).all()

    def test__4x4_mask_with_one_pixel__4x4_sub_grid__grid(self):
        mask = np.array([[True, True, True, True],
                         [True, False, False, True],
                         [True, False, False, True],
                         [True, True, True, False]])

        sub_grid = util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(mask=mask, pixel_scales=(2.0, 2.0),
                                                                                    sub_grid_size=4)

        sub_grid = np.round(sub_grid, decimals=1)

        assert (sub_grid == np.array([[1.6, -1.6], [1.6, -1.2], [1.6, -0.8], [1.6, -0.4],
                                      [1.2, -1.6], [1.2, -1.2], [1.2, -0.8], [1.2, -0.4],
                                      [0.8, -1.6], [0.8, -1.2], [0.8, -0.8], [0.8, -0.4],
                                      [0.4, -1.6], [0.4, -1.2], [0.4, -0.8], [0.4, -0.4],
                                      [1.6, 0.4], [1.6, 0.8], [1.6, 1.2], [1.6, 1.6],
                                      [1.2, 0.4], [1.2, 0.8], [1.2, 1.2], [1.2, 1.6],
                                      [0.8, 0.4], [0.8, 0.8], [0.8, 1.2], [0.8, 1.6],
                                      [0.4, 0.4], [0.4, 0.8], [0.4, 1.2], [0.4, 1.6],
                                      [-0.4, -1.6], [-0.4, -1.2], [-0.4, -0.8], [-0.4, -0.4],
                                      [-0.8, -1.6], [-0.8, -1.2], [-0.8, -0.8], [-0.8, -0.4],
                                      [-1.2, -1.6], [-1.2, -1.2], [-1.2, -0.8], [-1.2, -0.4],
                                      [-1.6, -1.6], [-1.6, -1.2], [-1.6, -0.8], [-1.6, -0.4],
                                      [-0.4, 0.4], [-0.4, 0.8], [-0.4, 1.2], [-0.4, 1.6],
                                      [-0.8, 0.4], [-0.8, 0.8], [-0.8, 1.2], [-0.8, 1.6],
                                      [-1.2, 0.4], [-1.2, 0.8], [-1.2, 1.2], [-1.2, 1.6],
                                      [-1.6, 0.4], [-1.6, 0.8], [-1.6, 1.2], [-1.6, 1.6],
                                      [-2.4, 2.4], [-2.4, 2.8], [-2.4, 3.2], [-2.4, 3.6],
                                      [-2.8, 2.4], [-2.8, 2.8], [-2.8, 3.2], [-2.8, 3.6],
                                      [-3.2, 2.4], [-3.2, 2.8], [-3.2, 3.2], [-3.2, 3.6],
                                      [-3.6, 2.4], [-3.6, 2.8], [-3.6, 3.2], [-3.6, 3.6]])).all()

    def test__4x3_mask_with_one_pixel__2x2_sub_grid__grid(self):
        mask = np.array([[True, True, True],
                         [True, False, True],
                         [True, False, False],
                         [False, True, True]])

        sub_grid = util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(mask=mask, pixel_scales=(3.0, 3.0),
                                                                                    sub_grid_size=2)

        assert (sub_grid == np.array([[2., -0.5], [2., 0.5], [1., -0.5], [1., 0.5],
                                      [-1., -0.5], [-1., 0.5], [-2., -0.5], [-2., 0.5],
                                       [-1., 2.5], [-1., 3.5], [-2., 2.5], [-2., 3.5],
                                      [-4., -3.5], [-4., -2.5], [-5., -3.5], [-5., -2.5]])).all()

    def test__3x4_mask_with_one_pixel__2x2_sub_grid__grid(self):

        mask = np.array([[True, True, True, False],
                         [True, False, False, True],
                         [False, True, False, True]])

        sub_grid = util.sub_grid_1d_masked_from_mask_pixel_scales_and_sub_grid_size(mask=mask, pixel_scales=(3.0, 3.0),
                                                                                    sub_grid_size=2)

        assert (sub_grid == np.array([[3.5, 4.], [3.5, 5.], [2.5, 4.], [2.5, 5.],
                                      [0.5, -2.], [0.5, -1.], [-0.5, -2.], [-0.5, -1.],
                                      [0.5, 1.], [0.5, 2.], [-0.5, 1.], [-0.5, 2.],
                                      [-2.5, -5.], [-2.5, -4.], [-3.5, -5.], [-3.5, -4.],
                                      [-2.5, 1.], [-2.5, 2.], [-3.5, 1.], [-3.5, 2.]])).all()


class TestGridPixelArcSecondConversion(object):

    def test__1d_arc_second_grid_to_1d_pixel_centred_grid__coordinates_in_centres_of_pixels(self):

        grid_arc_seconds = np.array([[1.0, -2.0], [1.0, 2.0],
                                     [-1.0, -2.0], [-1.0, 2.0]])

        grid_pixels = util.grid_arc_seconds_1d_to_grid_pixel_centres_1d(grid_arc_seconds=grid_arc_seconds, shape=(2, 2),
                                                                        pixel_scales=(2.0, 4.0))

        assert (grid_pixels == np.array([[0, 0], [0, 1],
                                         [1, 0], [1, 1]])).all()

        grid_arc_seconds = np.array([[3.0, -6.0], [3.0, 0.0], [3.0, 6.0],
                                         [0.0, -6.0], [0.0, 0.0], [0.0, 6.0],
                                         [-3.0, -6.0], [-3.0, 0.0], [-3.0, 6.0]])

        grid_pixels = util.grid_arc_seconds_1d_to_grid_pixel_centres_1d(grid_arc_seconds=grid_arc_seconds, shape=(3, 3),
                                                                        pixel_scales=(3.0, 6.0))

        assert (grid_pixels == np.array([[0, 0], [0, 1], [0, 2],
                                         [1, 0], [1, 1], [1, 2],
                                         [2, 0], [2, 1], [2, 2]])).all()

    def test__same_as_above_but_coordinates_are_top_left_of_each_pixel(self):

        grid_arc_seconds = np.array([[1.99, -3.99], [1.99, 0.01],
                                     [-0.01, -3.99], [-0.01, 0.01]])

        grid_pixels = util.grid_arc_seconds_1d_to_grid_pixel_centres_1d(grid_arc_seconds=grid_arc_seconds, shape=(2, 2),
                                                                        pixel_scales=(2.0, 4.0))

        assert (grid_pixels == np.array([[0, 0], [0, 1],
                                         [1, 0], [1, 1]])).all()

        grid_arc_seconds = np.array([[4.49, -8.99], [4.49, -2.99], [4.49, 3.01],
                                     [1.49, -8.99], [1.49, -2.99], [1.49, 3.01],
                                     [-1.51, -8.99], [-1.51, -2.99], [-1.51, 3.01]])

        grid_pixels = util.grid_arc_seconds_1d_to_grid_pixel_centres_1d(grid_arc_seconds=grid_arc_seconds, shape=(3, 3),
                                                                        pixel_scales=(3.0, 6.0))

        assert (grid_pixels == np.array([[0, 0], [0, 1], [0, 2],
                                         [1, 0], [1, 1], [1, 2],
                                         [2, 0], [2, 1], [2, 2]])).all()

    def test__same_as_above_but_coordinates_are_bottom_right_of_each_pixel(self):

        grid_arc_seconds = np.array([[0.01, -0.01], [0.01, 3.99],
                                     [-1.99, -0.01], [-1.99, 3.99]])

        grid_pixels = util.grid_arc_seconds_1d_to_grid_pixel_centres_1d(grid_arc_seconds=grid_arc_seconds, shape=(2, 2),
                                                                        pixel_scales=(2.0, 4.0))

        assert (grid_pixels == np.array([[0, 0], [0, 1],
                                         [1, 0], [1, 1]])).all()

        grid_arc_seconds = np.array([[1.51, -3.01], [1.51, 2.99], [1.51, 8.99],
                                     [-1.49, -3.01], [-1.49, 2.99], [-1.49, 8.99],
                                     [-4.49, -3.01], [-4.49, 2.99], [-4.49, 8.99]])

        grid_pixels = util.grid_arc_seconds_1d_to_grid_pixel_centres_1d(grid_arc_seconds=grid_arc_seconds, shape=(3, 3),
                                                                        pixel_scales=(3.0, 6.0))

        assert (grid_pixels == np.array([[0, 0], [0, 1], [0, 2],
                                         [1, 0], [1, 1], [1, 2],
                                         [2, 0], [2, 1], [2, 2]])).all()

    def test__1d_arc_second_grid_to_1d_pixel_grid__coordinates_in_centres_of_pixels(self):

        grid_arc_seconds = np.array([[1.0, -2.0], [1.0, 2.0],
                                     [-1.0, -2.0], [-1.0, 2.0]])

        grid_pixels = util.grid_arc_seconds_1d_to_grid_pixels_1d(grid_arc_seconds=grid_arc_seconds, shape=(2, 2),
                                                                        pixel_scales=(2.0, 4.0))

        assert (grid_pixels == np.array([[0.5, 0.5], [0.5, 1.5],
                                         [1.5, 0.5], [1.5, 1.5]])).all()

        grid_arc_seconds = np.array([[3.0, -6.0], [3.0, 0.0], [3.0, 6.0],
                                         [0.0, -6.0], [0.0, 0.0], [0.0, 6.0],
                                         [-3.0, -6.0], [-3.0, 0.0], [-3.0, 6.0]])

        grid_pixels = util.grid_arc_seconds_1d_to_grid_pixels_1d(grid_arc_seconds=grid_arc_seconds, shape=(3, 3),
                                                                        pixel_scales=(3.0, 6.0))

        assert (grid_pixels == np.array([[0.5, 0.5], [0.5, 1.5], [0.5, 2.5],
                                         [1.5, 0.5], [1.5, 1.5], [1.5, 2.5],
                                         [2.5, 0.5], [2.5, 1.5], [2.5, 2.5]])).all()

    def test__same_as_above__pixels__but_coordinates_are_top_left_of_each_pixel(self):

        grid_arc_seconds = np.array([[2.0, -4], [2.0, 0.0],
                                     [0.0, -4], [0.0, 0.0]])

        grid_pixels = util.grid_arc_seconds_1d_to_grid_pixels_1d(grid_arc_seconds=grid_arc_seconds, shape=(2, 2),
                                                                        pixel_scales=(2.0, 4.0))

        assert (grid_pixels == np.array([[0, 0], [0, 1],
                                         [1, 0], [1, 1]])).all()

        grid_arc_seconds = np.array([[4.5, -9.0], [4.5, -3.0], [4.5, 3.0],
                                     [1.5, -9.0], [1.5, -3.0], [1.5, 3.0],
                                     [-1.5, -9.0], [-1.5, -3.0], [-1.5, 3.0]])

        grid_pixels = util.grid_arc_seconds_1d_to_grid_pixels_1d(grid_arc_seconds=grid_arc_seconds, shape=(3, 3),
                                                                        pixel_scales=(3.0, 6.0))

        assert (grid_pixels == np.array([[0, 0], [0, 1], [0, 2],
                                         [1, 0], [1, 1], [1, 2],
                                         [2, 0], [2, 1], [2, 2]])).all()

    def test__same_as_above___pixels__but_coordinates_are_bottom_right_of_each_pixel(self):

        grid_arc_seconds = np.array([[0.0, 0.0], [0.0, 4.0],
                                     [-2.0, 0.0], [-2.0, 4.0]])

        grid_pixels = util.grid_arc_seconds_1d_to_grid_pixels_1d(grid_arc_seconds=grid_arc_seconds, shape=(2, 2),
                                                                        pixel_scales=(2.0, 4.0))

        assert (grid_pixels == np.array([[1, 1], [1, 2],
                                         [2, 1], [2, 2]])).all()

        grid_arc_seconds = np.array([[1.5, -3.0], [1.5, 3.0], [1.5, 9.0],
                                     [-1.5, -3.0], [-1.5, 3.0], [-1.5, 9.0],
                                     [-4.5, -3.0], [-4.5, 3.0], [-4.5, 9.0]])

        grid_pixels = util.grid_arc_seconds_1d_to_grid_pixels_1d(grid_arc_seconds=grid_arc_seconds, shape=(3, 3),
                                                                        pixel_scales=(3.0, 6.0))

        assert (grid_pixels == np.array([[1, 1], [1, 2], [1, 3],
                                         [2, 1], [2, 2], [2, 3],
                                         [3, 1], [3, 2], [3, 3]])).all()

    def test__1d_pixel_centre_grid_to_1d_arc_second_grid__coordinates_in_centres_of_pixels(self):

        grid_pixels = np.array([[0.5, 0.5], [0.5, 1.5],
                                 [1.5, 0.5], [1.5, 1.5]])

        grid_arc_seconds = util.grid_pixels_1d_to_grid_arc_seconds_1d(grid_pixels=grid_pixels, shape=(2, 2),
                                                                      pixel_scales=(2.0, 4.0))

        assert (grid_arc_seconds == np.array([[1.0, -2.0], [1.0, 2.0],
                                              [-1.0, -2.0], [-1.0, 2.0]])).all()

        grid_pixels = np.array([[0.5, 0.5], [0.5, 1.5], [0.5, 2.5],
                                 [1.5, 0.5], [1.5, 1.5], [1.5, 2.5],
                                 [2.5, 0.5], [2.5, 1.5], [2.5, 2.5]])

        grid_arc_seconds = util.grid_pixels_1d_to_grid_arc_seconds_1d(grid_pixels=grid_pixels, shape=(3, 3),
                                                                      pixel_scales=(3.0, 6.0))

        assert (grid_arc_seconds == np.array([[3.0, -6.0], [3.0, 0.0], [3.0, 6.0],
                                              [0.0, -6.0], [0.0, 0.0], [0.0, 6.0],
                                              [-3.0, -6.0], [-3.0, 0.0], [-3.0, 6.0]])).all()

    def test__same_as_above__pixel_to_arcsec__but_coordinates_are_top_left_of_each_pixel(self):

        grid_pixels = np.array([[0, 0], [0, 1],
                                 [1, 0], [1, 1]])

        grid_arc_seconds = util.grid_pixels_1d_to_grid_arc_seconds_1d(grid_pixels=grid_pixels, shape=(2, 2),
                                                                      pixel_scales=(2.0, 4.0))

        assert (grid_arc_seconds == np.array([[2.0, -4], [2.0, 0.0],
                                             [0.0, -4], [0.0, 0.0]])).all()



        grid_pixels = np.array([[0, 0], [0, 1], [0, 2],
                                 [1, 0], [1, 1], [1, 2],
                                 [2, 0], [2, 1], [2, 2]])

        grid_arc_seconds = util.grid_pixels_1d_to_grid_arc_seconds_1d(grid_pixels=grid_pixels, shape=(3, 3),
                                                                      pixel_scales=(3.0, 6.0))

        assert (grid_arc_seconds == np.array([[4.5, -9.0], [4.5, -3.0], [4.5, 3.0],
                                             [1.5, -9.0], [1.5, -3.0], [1.5, 3.0],
                                             [-1.5, -9.0], [-1.5, -3.0], [-1.5, 3.0]])).all()

    def test__same_as_above__pixel_to_arcsec_but_coordinates_are_bottom_right_of_each_pixel(self):

        grid_pixels = np.array([[1, 1], [1, 2],
                                [2, 1], [2, 2]])

        grid_arc_seconds = util.grid_pixels_1d_to_grid_arc_seconds_1d(grid_pixels=grid_pixels, shape=(2, 2),
                                                                      pixel_scales=(2.0, 4.0))

        assert (grid_arc_seconds == np.array([[0.0, 0.0], [0.0, 4.0],
                                     [-2.0, 0.0], [-2.0, 4.0]])).all()

        grid_pixels = np.array([[1, 1], [1, 2], [1, 3],
                                [2, 1], [2, 2], [2, 3],
                                [3, 1], [3, 2], [3, 3]])

        grid_arc_seconds = util.grid_pixels_1d_to_grid_arc_seconds_1d(grid_pixels=grid_pixels, shape=(3, 3),
                                                                      pixel_scales=(3.0, 6.0))

        assert (grid_arc_seconds == np.array([[1.5, -3.0], [1.5, 3.0], [1.5, 9.0],
                                     [-1.5, -3.0], [-1.5, 3.0], [-1.5, 9.0],
                                     [-4.5, -3.0], [-4.5, 3.0], [-4.5, 9.0]])).all()


class TestGridToPixel(object):

    def test__setup_3x3_image_one_pixel(self):
        mask = np.array([[True, True, True],
                         [True, False, True],
                         [True, True, True]])

        grid_to_pixel = util.grid_to_pixel_from_mask(mask)

        assert (grid_to_pixel == np.array([[1, 1]])).all()

    def test__setup_3x3_image__five_pixels(self):
        mask = np.array([[True, False, True],
                         [False, False, False],
                         [True, False, True]])

        grid_to_pixel = util.grid_to_pixel_from_mask(mask)

        assert (grid_to_pixel == np.array([        [0, 1],
                                           [1, 0], [1, 1], [1, 2],
                                                   [2, 1]])).all()

    def test__setup_3x4_image__six_pixels(self):

        mask = np.array([[True, False, True, True],
                         [False, False, False, True],
                         [True, False, True, False]])

        grid_to_pixel = util.grid_to_pixel_from_mask(mask)

        assert (grid_to_pixel == np.array([        [0, 1],
                                           [1, 0], [1, 1], [1, 2],
                                                   [2, 1],         [2, 3]])).all()

    def test__setup_4x3_image__six_pixels(self):
        mask = np.array([[True, False, True],
                         [False, False, False],
                         [True, False, True],
                         [True, True, False]])

        grid_to_pixel = util.grid_to_pixel_from_mask(mask)

        assert (grid_to_pixel == np.array([        [0, 1],
                                           [1, 0], [1, 1], [1, 2],
                                                   [2, 1],
                                                           [3, 2]])).all()


class TestSubToImage(object):

    def test__3x3_mask_with_1_pixel__2x2_sub_grid__correct_sub_to_image(self):
        mask = np.array([[True, True, True],
                         [True, False, True],
                         [True, True, True]])

        sub_to_image = util.sub_to_image_from_mask(mask, sub_grid_size=2)

        assert (sub_to_image == np.array([0, 0, 0, 0])).all()

    def test__3x3_mask_with_row_of_pixels_pixel__2x2_sub_grid__correct_sub_to_image(self):
        mask = np.array([[True, True, True],
                         [False, False, False],
                         [True, True, True]])

        sub_to_image = util.sub_to_image_from_mask(mask, sub_grid_size=2)

        assert (sub_to_image == np.array([0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2])).all()

    def test__3x3_mask_with_row_of_pixels_pixel__3x3_sub_grid__correct_sub_to_image(self):
        mask = np.array([[True, True, True],
                         [False, False, False],
                         [True, True, True]])

        sub_to_image = util.sub_to_image_from_mask(mask, sub_grid_size=3)

        assert (sub_to_image == np.array([0, 0, 0, 0, 0, 0, 0, 0, 0,
                                          1, 1, 1, 1, 1, 1, 1, 1, 1,
                                          2, 2, 2, 2, 2, 2, 2, 2, 2])).all()


class TestImageBorderPixels(object):

    def test__7x7_mask_one_central_pixel__is_entire_border(self):
        mask = np.array([[True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True],
                         [True, True, True, False, True, True, True],
                         [True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True]])

        border_pixels = util.border_pixels_from_mask(mask)

        assert (border_pixels == np.array([0])).all()

    def test__7x7_mask_nine_central_pixels__is_border(self):
        mask = np.array([[True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True],
                         [True, True, False, False, False, True, True],
                         [True, True, False, False, False, True, True],
                         [True, True, False, False, False, True, True],
                         [True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True]])

        border_pixels = util.border_pixels_from_mask(mask)

        assert (border_pixels == np.array([0, 1, 2, 3, 5, 6, 7, 8])).all()

    def test__7x7_mask_rectangle_of_fifteen_central_pixels__is_border(self):
        mask = np.array([[True, True, True, True, True, True, True],
                         [True, True, False, False, False, True, True],
                         [True, True, False, False, False, True, True],
                         [True, True, False, False, False, True, True],
                         [True, True, False, False, False, True, True],
                         [True, True, False, False, False, True, True],
                         [True, True, True, True, True, True, True]])

        border_pixels = util.border_pixels_from_mask(mask)

        assert (border_pixels == np.array([0, 1, 2, 3, 5, 6, 8, 9, 11, 12, 13, 14])).all()

    def test__8x7_mask_add_edge_pixels__also_in_border(self):
        mask = np.array([[True, True, True, True, True, True, True],
                         [True, True, True, False, True, True, True],
                         [True, True, False, False, False, True, True],
                         [True, True, False, False, False, True, True],
                         [True, False, False, False, False, False, True],
                         [True, True, False, False, False, True, True],
                         [True, True, False, False, False, True, True],
                         [True, True, True, True, True, True, True]])

        border_pixels = util.border_pixels_from_mask(mask)

        assert (border_pixels == np.array([0, 1, 2, 3, 4, 6, 7, 8, 10, 11, 12, 14, 15, 16, 17])).all()

    def test__8x7_mask_big_square(self):
        mask = np.array([[True, True, True, True, True, True, True],
                         [True, False, False, False, False, False, True],
                         [True, False, False, False, False, False, True],
                         [True, False, False, False, False, False, True],
                         [True, False, False, False, False, False, True],
                         [True, False, False, False, False, False, True],
                         [True, False, False, False, False, False, True],
                         [True, True, True, True, True, True, True]])

        border_pixels = util.border_pixels_from_mask(mask)

        assert (border_pixels == np.array([0, 1, 2, 3, 4, 5, 9, 10, 14, 15, 19, 20, 24, 25, 26, 27, 28, 29])).all()

    def test__7x8_mask_add_edge_pixels__also_in_border(self):
        mask = np.array([[True, True, True, True, True, True, True, True],
                         [True, True, True, False, True, True, True, True],
                         [True, True, False, False, False, True, True, True],
                         [True, True, False, False, False, True, True, True],
                         [True, False, False, False, False, False, True, True],
                         [True, True, False, False, False, True, True, True],
                         [True, True, True, True, True, True, True, True]])

        border_pixels = util.border_pixels_from_mask(mask)

        assert (border_pixels == np.array([0, 1, 2, 3, 4, 6, 7, 8, 10, 11, 12, 13, 14])).all()

    def test__7x8_mask_big_square(self):
        mask = np.array([[True, True, True, True, True, True, True, True],
                         [True, False, False, False, False, False, True, True],
                         [True, False, False, False, False, False, True, True],
                         [True, False, False, False, False, False, True, True],
                         [True, False, False, False, False, False, True, True],
                         [True, False, False, False, False, False, True, True],
                         [True, True, True, True, True, True, True, True]])

        border_pixels = util.border_pixels_from_mask(mask)

        assert (border_pixels == np.array([0, 1, 2, 3, 4, 5, 9, 10, 14, 15, 19, 20, 21, 22, 23, 24])).all()


class TestSubBorderPixels(object):

    def test__7x7_mask__2x2_sub_grid__nine_central_pixels__is_border(self):
        mask = np.array([[True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True],
                         [True, True, False, False, False, True, True],
                         [True, True, False, False, False, True, True],
                         [True, True, False, False, False, True, True],
                         [True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True]])

        border_sub_pixels = util.border_sub_pixels_from_mask_pixel_scales_and_sub_grid_size(mask=mask,
                                                                                            pixel_scales=(3.0, 3.0),
                                                                                            sub_grid_size=2)

        assert (border_sub_pixels == np.array([0, 4, 9, 12, 21, 26, 30, 35])).all()

    def test__7x7_mask__4x4_sub_grid_nine_central_pixels__is_border(self):
        mask = np.array([[True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True],
                         [True, True, False, False, False, True, True],
                         [True, True, False, False, False, True, True],
                         [True, True, False, False, False, True, True],
                         [True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True]])

        border_sub_pixels = util.border_sub_pixels_from_mask_pixel_scales_and_sub_grid_size(mask=mask,
                                                                                            pixel_scales=(3.0, 3.0),
                                                                                            sub_grid_size=4)

        assert (border_sub_pixels == np.array([0, 16, 35, 48, 83, 108, 124, 143])).all()

    def test__7x7_mask_rectangle_of_fifteen_central_pixels__is_border(self):
        mask = np.array([[True, True, True, True, True, True, True],
                         [True, True, False, False, False, True, True],
                         [True, True, False, False, False, True, True],
                         [True, True, False, False, False, True, True],
                         [True, True, False, False, False, True, True],
                         [True, True, False, False, False, True, True],
                         [True, True, True, True, True, True, True]])

        border_sub_pixels = util.border_sub_pixels_from_mask_pixel_scales_and_sub_grid_size(mask=mask,
                                                                                            pixel_scales=(3.0, 3.0),
                                                                                            sub_grid_size=2)
        assert (border_sub_pixels == np.array([0, 4, 9, 12, 21, 24, 33, 38, 47, 50, 54, 59])).all()

    def test__8x7_mask_add_edge_pixels__also_in_border(self):
        mask = np.array([[True, True, True, True, True, True, True],
                         [True, True, True, False, True, True, True],
                         [True, True, False, False, False, True, True],
                         [True, True, False, False, False, True, True],
                         [True, False, False, False, False, False, True],
                         [True, True, False, False, False, True, True],
                         [True, True, False, False, False, True, True],
                         [True, True, True, True, True, True, True]])

        border_sub_pixels = util.border_sub_pixels_from_mask_pixel_scales_and_sub_grid_size(mask=mask,
                                                                                            pixel_scales=(3.0, 3.0),
                                                                                            sub_grid_size=2)
        assert (border_sub_pixels == np.array([0, 4, 8, 13, 16, 25, 30, 34, 43, 47, 50, 59, 62, 66, 71])).all()

    def test__7x8_mask_add_edge_pixels__also_in_border(self):
        mask = np.array([[True, True, True, True, True, True, True, True],
                         [True, True, True, False, True, True, True, True],
                         [True, True, False, False, False, True, True, True],
                         [True, True, False, False, False, True, True, True],
                         [True, False, False, False, False, False, True, True],
                         [True, True, False, False, False, True, True, True],
                         [True, True, True, True, True, True, True, True]])

        border_sub_pixels = util.border_sub_pixels_from_mask_pixel_scales_and_sub_grid_size(mask=mask,
                                                                                            pixel_scales=(3.0, 3.0),
                                                                                            sub_grid_size=2)
        assert (border_sub_pixels == np.array([0, 4, 8, 13, 16, 25, 30, 34, 43, 47, 50, 54, 59])).all()


class TestMaskCircular(object):

    def test__input_big_mask__mask(self):
        mask = util.mask_circular_from_shape_pixel_scale_and_radius(shape=(3, 3), pixel_scale=1.0,
                                                                    radius_arcsec=5.0)

        assert mask.shape == (3, 3)
        assert (mask == np.array([[False, False, False],
                                  [False, False, False],
                                  [False, False, False]])).all()

    def test__3x3_mask_input_radius_small__mask(self):
        mask = util.mask_circular_from_shape_pixel_scale_and_radius(shape=(3, 3), pixel_scale=1.0,
                                                                    radius_arcsec=0.5)

        assert (mask == np.array([[True, True, True],
                                  [True, False, True],
                                  [True, True, True]])).all()

    def test__3x3_mask_input_radius_medium__mask(self):
        mask = util.mask_circular_from_shape_pixel_scale_and_radius(shape=(3, 3), pixel_scale=1.0,
                                                                    radius_arcsec=1.3)

        assert (mask == np.array([[True,  False, True],
                                  [False, False, False],
                                  [True,  False, True]])).all()

    def test__3x3_mask_input_radius_large__mask(self):
        mask = util.mask_circular_from_shape_pixel_scale_and_radius(shape=(3, 3), pixel_scale=1.0,
                                                                    radius_arcsec=3.0)

        assert (mask == np.array([[False, False, False],
                                  [False, False, False],
                                  [False, False, False]])).all()

    def test__4x3_mask_input_radius_small__mask(self):
        mask = util.mask_circular_from_shape_pixel_scale_and_radius(shape=(4, 3), pixel_scale=1.0,
                                                                    radius_arcsec=0.5)

        assert (mask == np.array([[True, True, True],
                                  [True, False, True],
                                  [True, False, True],
                                  [True, True, True]])).all()

    def test__4x3_mask_input_radius_medium__mask(self):
        mask = util.mask_circular_from_shape_pixel_scale_and_radius(shape=(4, 3), pixel_scale=1.0,
                                                                    radius_arcsec=1.5001)

        assert (mask == np.array([[True, False, True],
                                  [False, False, False],
                                  [False, False, False],
                                  [True, False, True]])).all()

    def test__4x3_mask_input_radius_large__mask(self):
        mask = util.mask_circular_from_shape_pixel_scale_and_radius(shape=(4, 3), pixel_scale=1.0,
                                                                    radius_arcsec=3.0)

        assert (mask == np.array([[False, False, False],
                                  [False, False, False],
                                  [False, False, False],
                                  [False, False, False]])).all()

    def test__4x4_mask_input_radius_small__mask(self):
        mask = util.mask_circular_from_shape_pixel_scale_and_radius(shape=(4, 4), pixel_scale=1.0,
                                                                    radius_arcsec=0.72)

        assert (mask == np.array([[True, True, True, True],
                                  [True, False, False, True],
                                  [True, False, False, True],
                                  [True, True, True, True]])).all()

    def test__4x4_mask_input_radius_medium__mask(self):
        mask = util.mask_circular_from_shape_pixel_scale_and_radius(shape=(4, 4), pixel_scale=1.0,
                                                                    radius_arcsec=1.7)

        assert (mask == np.array([[True, False, False, True],
                                  [False, False, False, False],
                                  [False, False, False, False],
                                  [True, False, False, True]])).all()

    def test__4x4_mask_input_radius_large__mask(self):
        mask = util.mask_circular_from_shape_pixel_scale_and_radius(shape=(4, 4), pixel_scale=1.0,
                                                                    radius_arcsec=3.0)

        assert (mask == np.array([[False, False, False, False],
                                  [False, False, False, False],
                                  [False, False, False, False],
                                  [False, False, False, False]])).all()

    def test__centre_shift__simple_shift_downwards(self):
        mask = util.mask_circular_from_shape_pixel_scale_and_radius(shape=(3, 3), pixel_scale=3.0,
                                                                    radius_arcsec=0.5, centre=(-3, 0))

        assert mask.shape == (3, 3)
        assert (mask == np.array([[True, True, True],
                                  [True, True, True],
                                  [True, False, True]])).all()

    def test__centre_shift__simple_shift_right(self):
        mask = util.mask_circular_from_shape_pixel_scale_and_radius(shape=(3, 3), pixel_scale=3.0,
                                                                    radius_arcsec=0.5, centre=(0.0, 3.0))

        assert mask.shape == (3, 3)
        assert (mask == np.array([[True, True, True],
                                  [True, True, False],
                                  [True, True, True]])).all()

    def test__centre_shift__diagonal_shift(self):
        mask = util.mask_circular_from_shape_pixel_scale_and_radius(shape=(3, 3), pixel_scale=3.0,
                                                                    radius_arcsec=0.5, centre=(3, 3))

        assert (mask == np.array([[True, True, False],
                                  [True, True, True],
                                  [True, True, True]])).all()


class TestMaskAnnular(object):

    def test__3x3_mask_inner_radius_zero_outer_radius_small__mask(self):
        mask = util.mask_annular_from_shape_pixel_scale_and_radii(shape=(3, 3), pixel_scale=1.0,
                                                                  inner_radius_arcsec=0.0, outer_radius_arcsec=0.5)

        assert (mask == np.array([[True, True, True],
                                  [True, False, True],
                                  [True, True, True]])).all()

    def test__3x3_mask_inner_radius_small_outer_radius_large__mask(self):
        mask = util.mask_annular_from_shape_pixel_scale_and_radii(shape=(3, 3), pixel_scale=1.0,
                                                                  inner_radius_arcsec=0.5, outer_radius_arcsec=3.0)

        assert (mask == np.array([[False, False, False],
                                  [False, True, False],
                                  [False, False, False]])).all()

    def test__4x3_mask_inner_radius_small_outer_radius_medium__mask(self):
        mask = util.mask_annular_from_shape_pixel_scale_and_radii(shape=(4, 3), pixel_scale=1.0,
                                                                  inner_radius_arcsec=0.51, outer_radius_arcsec=1.51)

        assert (mask == np.array([[True, False, True],
                                  [False, True, False],
                                  [False, True, False],
                                  [True, False, True]])).all()

    def test__4x3_mask_inner_radius_medium_outer_radius_large__mask(self):
        mask = util.mask_annular_from_shape_pixel_scale_and_radii(shape=(4, 3), pixel_scale=1.0,
                                                                  inner_radius_arcsec=1.51, outer_radius_arcsec=3.0)

        assert (mask == np.array([[False, True, False],
                                  [True, True, True],
                                  [True, True, True],
                                  [False, True, False]])).all()

    def test__3x3_mask_inner_radius_small_outer_radius_medium__mask(self):
        mask = util.mask_annular_from_shape_pixel_scale_and_radii(shape=(4, 4), pixel_scale=1.0,
                                                                  inner_radius_arcsec=0.81, outer_radius_arcsec=2.0)

        assert (mask == np.array([[True, False, False, True],
                                  [False, True, True, False],
                                  [False, True, True, False],
                                  [True, False, False, True]])).all()

    def test__4x4_mask_inner_radius_medium_outer_radius_large__mask(self):
        mask = util.mask_annular_from_shape_pixel_scale_and_radii(shape=(4, 4), pixel_scale=1.0,
                                                                  inner_radius_arcsec=1.71, outer_radius_arcsec=3.0)

        assert (mask == np.array([[False, True, True, False],
                                  [True, True, True, True],
                                  [True, True, True, True],
                                  [False, True, True, False]])).all()

    def test__centre_shift__simple_shift_upwards(self):
        mask = util.mask_annular_from_shape_pixel_scale_and_radii(shape=(3, 3), pixel_scale=3.0,
                                                                  inner_radius_arcsec=0.5,
                                                                  outer_radius_arcsec=9.0, centre=(3.0, 0.0))

        assert mask.shape == (3, 3)
        assert (mask == np.array([[False, True, False],
                                  [False, False, False],
                                  [False, False, False]])).all()

    def test__centre_shift__simple_shift_forward(self):
        mask = util.mask_annular_from_shape_pixel_scale_and_radii(shape=(3, 3), pixel_scale=3.0,
                                                                  inner_radius_arcsec=0.5,
                                                                  outer_radius_arcsec=9.0, centre=(0.0, 3.0))

        assert mask.shape == (3, 3)
        assert (mask == np.array([[False, False, False],
                                  [False, False, True],
                                  [False, False, False]])).all()

    def test__centre_shift__diagonal_shift(self):
        mask = util.mask_annular_from_shape_pixel_scale_and_radii(shape=(3, 3), pixel_scale=3.0,
                                                                  inner_radius_arcsec=0.5,
                                                                  outer_radius_arcsec=9.0, centre=(-3.0, 3.0))

        assert mask.shape == (3, 3)
        assert (mask == np.array([[False, False, False],
                                  [False, False, False],
                                  [False, False, True]])).all()


class TestMaskAntiAnnular(object):

    def test__5x5_mask_inner_radius_includes_central_pixel__outer_extended_beyond_radius(self):

        mask = util.mask_anti_annular_from_shape_pixel_scale_and_radii(shape=(5, 5), pixel_scale=1.0,
                                                                       inner_radius_arcsec=0.5, outer_radius_arcsec=10.0,
                                                                       outer_radius_2_arcsec=20.0)

        assert (mask == np.array([[True, True, True, True, True],
                                  [True, True, True, True, True],
                                  [True, True, False, True, True],
                                  [True, True, True, True, True],
                                  [True, True, True, True, True]])).all()

    def test__5x5_mask_inner_radius_includes_3x3_central_pixels__outer_extended_beyond_radius(self):

        mask = util.mask_anti_annular_from_shape_pixel_scale_and_radii(shape=(5, 5), pixel_scale=1.0,
                                                                       inner_radius_arcsec=1.5, outer_radius_arcsec=10.0,
                                                                       outer_radius_2_arcsec=20.0)

        assert (mask == np.array([[True,  True,  True,  True, True],
                                  [True, False, False, False, True],
                                  [True, False, False, False, True],
                                  [True, False, False, False, True],
                                  [True,  True,  True,  True, True]])).all()

    def test__5x5_mask_inner_radius_includes_central_pixel__outer_radius_includes_outer_pixels(self):

        mask = util.mask_anti_annular_from_shape_pixel_scale_and_radii(shape=(5, 5), pixel_scale=1.0,
                                                                       inner_radius_arcsec=0.5, outer_radius_arcsec=1.5,
                                                                       outer_radius_2_arcsec=20.0)

        assert (mask == np.array([[False, False, False, False, False],
                                  [False, True,  True,  True,  False],
                                  [False, True, False,  True,  False],
                                  [False, True,  True,  True,  False],
                                  [False, False, False, False, False]])).all()

    def test__7x7_second_outer_radius_mask_works_too(self):

        mask = util.mask_anti_annular_from_shape_pixel_scale_and_radii(shape=(7, 7), pixel_scale=1.0,
                                                                       inner_radius_arcsec=0.5, outer_radius_arcsec=1.5,
                                                                       outer_radius_2_arcsec=2.9)

        assert (mask == np.array([[True,  True,  True,  True,  True,  True, True],
                                  [True, False, False, False, False, False, True],
                                  [True, False, True,   True,  True, False, True],
                                  [True, False, True,  False,  True, False, True],
                                  [True, False, True,   True,  True, False, True],
                                  [True, False, False, False, False, False, True],
                                  [True,  True,  True,  True,  True,  True, True]])).all()

    def test__centre_shift__diagonal_shift(self):

        mask = util.mask_anti_annular_from_shape_pixel_scale_and_radii(shape=(7, 7), pixel_scale=3.0,
                                                                       inner_radius_arcsec=1.5, outer_radius_arcsec=4.5,
                                                                       outer_radius_2_arcsec=8.7, centre=(3.0, -3.0))

        assert (mask == np.array([[True,  True,  True,  True,  True,  True,  True],
                                  [True,  True,  True,  True,  True,  True,  True],
                                  [True,  True, False, False, False, False, False],
                                  [True,  True, False, True,   True,  True, False],
                                  [True,  True, False, True,  False,  True, False],
                                  [True,  True, False, True,   True,  True, False],
                                  [True,  True, False, False, False, False, False]])).all()


class TestMaskBlurring(object):

    def test__size__3x3_small_mask(self):
        mask = np.array([[True, True, True],
                         [True, False, True],
                         [True, True, True]])

        blurring_mask = util.mask_blurring_from_mask_and_psf_shape(mask, psf_shape=(3, 3))

        assert (blurring_mask == np.array([[False, False, False],
                                           [False, True, False],
                                           [False, False, False]])).all()

    def test__size__3x3__large_mask(self):
        mask = np.array([[True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True],
                         [True, True, True, False, True, True, True],
                         [True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True]])

        blurring_mask = util.mask_blurring_from_mask_and_psf_shape(mask, psf_shape=(3, 3))

        assert (blurring_mask == np.array([[True, True, True, True, True, True, True],
                                           [True, True, True, True, True, True, True],
                                           [True, True, False, False, False, True, True],
                                           [True, True, False, True, False, True, True],
                                           [True, True, False, False, False, True, True],
                                           [True, True, True, True, True, True, True],
                                           [True, True, True, True, True, True, True]])).all()

    def test__size__5x5__large_mask(self):
        mask = np.array([[True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True],
                         [True, True, True, False, True, True, True],
                         [True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True]])

        blurring_mask = util.mask_blurring_from_mask_and_psf_shape(mask, psf_shape=(5, 5))

        assert (blurring_mask == np.array([[True, True, True, True, True, True, True],
                                           [True, False, False, False, False, False, True],
                                           [True, False, False, False, False, False, True],
                                           [True, False, False, True, False, False, True],
                                           [True, False, False, False, False, False, True],
                                           [True, False, False, False, False, False, True],
                                           [True, True, True, True, True, True, True]])).all()

    def test__size__5x3__large_mask(self):
        mask = np.array([[True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True],
                         [True, True, True, False, True, True, True],
                         [True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True]])

        blurring_mask = util.mask_blurring_from_mask_and_psf_shape(mask, psf_shape=(5, 3))

        assert (blurring_mask == np.rot90(np.array([[True, True, True, True, True, True, True],
                                                    [True, True, True, True, True, True, True],
                                                    [True, False, False, False, False, False, True],
                                                    [True, False, False, True, False, False, True],
                                                    [True, False, False, False, False, False, True],
                                                    [True, True, True, True, True, True, True],
                                                    [True, True, True, True, True, True, True]]))).all()

    def test__size__3x5__large_mask(self):
        mask = np.array([[True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True],
                         [True, True, True, False, True, True, True],
                         [True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True]])

        blurring_mask = util.mask_blurring_from_mask_and_psf_shape(mask, psf_shape=(3, 5))

        assert (blurring_mask == np.rot90(np.array([[True, True, True, True, True, True, True],
                                                    [True, True, False, False, False, True, True],
                                                    [True, True, False, False, False, True, True],
                                                    [True, True, False, True, False, True, True],
                                                    [True, True, False, False, False, True, True],
                                                    [True, True, False, False, False, True, True],
                                                    [True, True, True, True, True, True, True]]))).all()

    def test__size__3x3__multiple_points(self):
        mask = np.array([[True, True, True, True, True, True, True],
                         [True, False, True, True, True, False, True],
                         [True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True],
                         [True, False, True, True, True, False, True],
                         [True, True, True, True, True, True, True]])

        blurring_mask = util.mask_blurring_from_mask_and_psf_shape(mask, psf_shape=(3, 3))

        assert (blurring_mask == np.array([[False, False, False, True, False, False, False],
                                           [False, True, False, True, False, True, False],
                                           [False, False, False, True, False, False, False],
                                           [True, True, True, True, True, True, True],
                                           [False, False, False, True, False, False, False],
                                           [False, True, False, True, False, True, False],
                                           [False, False, False, True, False, False, False]])).all()

    def test__size__5x5__multiple_points(self):
        mask = np.array([[True, True, True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True, True, True],
                         [True, True, False, True, True, True, False, True, True],
                         [True, True, True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True, True, True],
                         [True, True, False, True, True, True, False, True, True],
                         [True, True, True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True, True, True]])

        blurring_mask = util.mask_blurring_from_mask_and_psf_shape(mask, psf_shape=(5, 5))

        assert (blurring_mask == np.array([[False, False, False, False, False, False, False, False, False],
                                           [False, False, False, False, False, False, False, False, False],
                                           [False, False, True, False, False, False, True, False, False],
                                           [False, False, False, False, False, False, False, False, False],
                                           [False, False, False, False, False, False, False, False, False],
                                           [False, False, False, False, False, False, False, False, False],
                                           [False, False, True, False, False, False, True, False, False],
                                           [False, False, False, False, False, False, False, False, False],
                                           [False, False, False, False, False, False, False, False,
                                            False]])).all()

    def test__size__5x3__multiple_points(self):
        mask = np.array([[True, True, True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True, True, True],
                         [True, True, False, True, True, True, False, True, True],
                         [True, True, True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True, True, True],
                         [True, True, False, True, True, True, False, True, True],
                         [True, True, True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True, True, True]])

        blurring_mask = util.mask_blurring_from_mask_and_psf_shape(mask, psf_shape=(5, 3))

        assert (blurring_mask == np.rot90(np.array([[True, True, True, True, True, True, True, True, True],
                                                    [False, False, False, False, False, False, False, False, False],
                                                    [False, False, True, False, False, False, True, False, False],
                                                    [False, False, False, False, False, False, False, False, False],
                                                    [True, True, True, True, True, True, True, True, True],
                                                    [False, False, False, False, False, False, False, False, False],
                                                    [False, False, True, False, False, False, True, False, False],
                                                    [False, False, False, False, False, False, False, False, False],
                                                    [True, True, True, True, True, True, True, True, True]]))).all()

    def test__size__3x5__multiple_points(self):
        mask = np.array([[True, True, True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True, True, True],
                         [True, True, False, True, True, True, False, True, True],
                         [True, True, True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True, True, True],
                         [True, True, False, True, True, True, False, True, True],
                         [True, True, True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True, True, True]])

        blurring_mask = util.mask_blurring_from_mask_and_psf_shape(mask, psf_shape=(3, 5))

        assert (blurring_mask == np.rot90(np.array([[True, False, False, False, True, False, False, False, True],
                                                    [True, False, False, False, True, False, False, False, True],
                                                    [True, False, True, False, True, False, True, False, True],
                                                    [True, False, False, False, True, False, False, False, True],
                                                    [True, False, False, False, True, False, False, False, True],
                                                    [True, False, False, False, True, False, False, False, True],
                                                    [True, False, True, False, True, False, True, False, True],
                                                    [True, False, False, False, True, False, False, False, True],
                                                    [True, False, False, False, True, False, False, False,
                                                     True]]))).all()

    def test__size__3x3__even_sized_image(self):
        mask = np.array([[True, True, True, True, True, True, True, True],
                         [True, False, True, True, True, False, True, True],
                         [True, True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True, True],
                         [True, False, True, True, True, False, True, True],
                         [True, True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True, True]])

        blurring_mask = util.mask_blurring_from_mask_and_psf_shape(mask, psf_shape=(3, 3))

        assert (blurring_mask == np.array([[False, False, False, True, False, False, False, True],
                                           [False, True, False, True, False, True, False, True],
                                           [False, False, False, True, False, False, False, True],
                                           [True, True, True, True, True, True, True, True],
                                           [False, False, False, True, False, False, False, True],
                                           [False, True, False, True, False, True, False, True],
                                           [False, False, False, True, False, False, False, True],
                                           [True, True, True, True, True, True, True, True]])).all()

    def test__size__5x5__even_sized_image(self):
        mask = np.array([[True, True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True, True],
                         [True, True, True, True, True, False, True, True],
                         [True, True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True, True]])

        blurring_mask = util.mask_blurring_from_mask_and_psf_shape(mask, psf_shape=(5, 5))

        assert (blurring_mask == np.array([[True, True, True, True, True, True, True, True],
                                           [True, True, True, True, True, True, True, True],
                                           [True, True, True, True, True, True, True, True],
                                           [True, True, True, False, False, False, False, False],
                                           [True, True, True, False, False, False, False, False],
                                           [True, True, True, False, False, True, False, False],
                                           [True, True, True, False, False, False, False, False],
                                           [True, True, True, False, False, False, False, False]])).all()

    def test__size__3x3__rectangular_8x9_image(self):
        mask = np.array([[True, True, True, True, True, True, True, True, True],
                         [True, False, True, True, True, False, True, True, True],
                         [True, True, True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True, True, True],
                         [True, False, True, True, True, False, True, True, True],
                         [True, True, True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True, True, True]])

        blurring_mask = util.mask_blurring_from_mask_and_psf_shape(mask, psf_shape=(3, 3))

        assert (blurring_mask == np.array([[False, False, False, True, False, False, False, True, True],
                                           [False, True, False, True, False, True, False, True, True],
                                           [False, False, False, True, False, False, False, True, True],
                                           [True, True, True, True, True, True, True, True, True],
                                           [False, False, False, True, False, False, False, True, True],
                                           [False, True, False, True, False, True, False, True, True],
                                           [False, False, False, True, False, False, False, True, True],
                                           [True, True, True, True, True, True, True, True, True]])).all()

    def test__size__3x3__rectangular_9x8_image(self):
        mask = np.array([[True, True, True, True, True, True, True, True],
                         [True, False, True, True, True, False, True, True],
                         [True, True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True, True],
                         [True, False, True, True, True, False, True, True],
                         [True, True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True, True]])

        blurring_mask = util.mask_blurring_from_mask_and_psf_shape(mask, psf_shape=(3, 3))

        assert (blurring_mask == np.array([[False, False, False, True, False, False, False, True],
                                           [False, True, False, True, False, True, False, True],
                                           [False, False, False, True, False, False, False, True],
                                           [True, True, True, True, True, True, True, True],
                                           [False, False, False, True, False, False, False, True],
                                           [False, True, False, True, False, True, False, True],
                                           [False, False, False, True, False, False, False, True],
                                           [True, True, True, True, True, True, True, True],
                                           [True, True, True, True, True, True, True, True]])).all()

    def test__size__5x5__multiple_points__mask_extends_beyond_border_so_raises_mask_exception(self):
        mask = np.array([[True, True, True, True, True, True, True],
                         [True, False, True, True, True, False, True],
                         [True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True],
                         [True, True, True, True, True, True, True],
                         [True, False, True, True, True, False, True],
                         [True, True, True, True, True, True, True]])

        with pytest.raises(exc.MaskException):
            util.mask_blurring_from_mask_and_psf_shape(mask, psf_shape=(5, 5))


class TestMap2DArrayTo1d(object):

    def test__setup_3x3_data(self):
        array_2d = np.array([[1, 2, 3],
                             [4, 5, 6],
                             [7, 8, 9]])

        mask = np.array([[True, True, True],
                         [True, False, True],
                         [True, True, True]])

        array_1d = util.map_2d_array_to_masked_1d_array_from_array_2d_and_mask(mask, array_2d)

        assert (array_1d == np.array([5])).all()

    def test__setup_3x3_array__five_now_in_mask(self):
        array_2d = np.array([[1, 2, 3],
                             [4, 5, 6],
                             [7, 8, 9]])

        mask = np.array([[True, False, True],
                         [False, False, False],
                         [True, False, True]])

        array_1d = util.map_2d_array_to_masked_1d_array_from_array_2d_and_mask(mask, array_2d)

        assert (array_1d == np.array([2, 4, 5, 6, 8])).all()

    def test__setup_3x4_array(self):
        array_2d = np.array([[1, 2, 3, 4],
                             [5, 6, 7, 8],
                             [9, 10, 11, 12]])

        mask = np.array([[True, False, True, True],
                         [False, False, False, True],
                         [True, False, True, False]])

        array_1d = util.map_2d_array_to_masked_1d_array_from_array_2d_and_mask(mask, array_2d)

        assert (array_1d == np.array([2, 5, 6, 7, 10, 12])).all()

    def test__setup_4x3_array__five_now_in_mask(self):
        array_2d = np.array([[1, 2, 3],
                             [4, 5, 6],
                             [7, 8, 9],
                             [10, 11, 12]])

        mask = np.array([[True, False, True],
                         [False, False, False],
                         [True, False, True],
                         [True, True, True]])

        array_1d = util.map_2d_array_to_masked_1d_array_from_array_2d_and_mask(mask, array_2d)

        assert (array_1d == np.array([2, 4, 5, 6, 8])).all()


class TestMapMasked1DArrayTo2d(object):

    def test__2d_array_is_2x2__is_not_masked__maps_correctly(self):
        array_1d = np.array([1.0, 2.0, 3.0, 4.0])

        one_to_two = np.array([[0, 0], [0, 1], [1, 0], [1, 1]])
        shape = (2, 2)

        array_2d = util.map_masked_1d_array_to_2d_array_from_array_1d_shape_and_one_to_two(array_1d, shape, one_to_two)

        assert (array_2d == np.array([[1.0, 2.0],
                                      [3.0, 4.0]])).all()

    def test__2d_array_is_2x2__is_masked__maps_correctly(self):
        array_1d = np.array([1.0, 2.0, 3.0])

        one_to_two = np.array([[0, 0], [0, 1], [1, 0]])
        shape = (2, 2)

        array_2d = util.map_masked_1d_array_to_2d_array_from_array_1d_shape_and_one_to_two(array_1d, shape, one_to_two)

        assert (array_2d == np.array([[1.0, 2.0],
                                      [3.0, 0.0]])).all()

    def test__different_shape_and_mappings(self):
        array_1d = np.array([1.0, 2.0, 3.0, -1.0, -2.0, -3.0])

        one_to_two = np.array([[0, 0], [0, 1], [1, 0], [2, 0], [2, 1], [2, 3]])
        shape = (3, 4)

        array_2d = util.map_masked_1d_array_to_2d_array_from_array_1d_shape_and_one_to_two(array_1d, shape, one_to_two)

        assert (array_2d == np.array([[1.0, 2.0, 0.0, 0.0],
                                      [3.0, 0.0, 0.0, 0.0],
                                      [-1.0, -2.0, 0.0, -3.0]])).all()


class TestMapUnmasked1dArrayTo2d(object):

    def test__1d_array_in__maps_it_to_4x4_2d_array(self):
        array_1d = np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0, 14.0, 15.0, 16.0])
        array_2d = util.map_unmasked_1d_array_to_2d_array_from_array_1d_and_shape(array_1d, shape=(4, 4))

        assert (array_2d == np.array([[1.0, 2.0, 3.0, 4.0],
                                      [5.0, 6.0, 7.0, 8.0],
                                      [9.0, 10.0, 11.0, 12.0],
                                      [13.0, 14.0, 15.0, 16.0]])).all()

    def test__1d_array_in__can_map_it_to_2x3_2d_array(self):
        array_1d = np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0])
        array_2d = util.map_unmasked_1d_array_to_2d_array_from_array_1d_and_shape(array_1d, shape=(2, 3))

        assert (array_2d == np.array([[1.0, 2.0, 3.0],
                                      [4.0, 5.0, 6.0]])).all()

    def test__1d_array_in__can_map_it_to_3x2_2d_array(self):
        array_1d = np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0])
        array_2d = util.map_unmasked_1d_array_to_2d_array_from_array_1d_and_shape(array_1d, shape=(3, 2))

        assert (array_2d == np.array([[1.0, 2.0],
                                      [3.0, 4.0],
                                      [5.0, 6.0]])).all()


class TestTrim:

    def test__from_even_to_odd_or_odd_to_even__raises_error(self):

        array = np.ones((5, 5))

        with pytest.raises(ValueError):
            util.trim_array_2d_around_centre(array_2d=array, new_shape=(4, 4))

        array = np.ones((6, 6))

        with pytest.raises(ValueError):
            util.trim_array_2d_around_centre(array_2d=array, new_shape=(3, 3))

        array = np.ones((6, 5))

        with pytest.raises(ValueError):
            util.trim_array_2d_around_centre(array_2d=array, new_shape=(3, 3))

        array = np.ones((5, 6))

        with pytest.raises(ValueError):
            util.trim_array_2d_around_centre(array_2d=array, new_shape=(3, 3))

    def test__from_7x7_to_3x3(self):
        array = np.ones((7, 7))
        array[3, 3] = 2.0

        modified = util.trim_array_2d_around_centre(array_2d=array, new_shape=(3, 3))

        assert (modified == np.array([[1.0, 1.0, 1.0],
                                      [1.0, 2.0, 1.0],
                                      [1.0, 1.0, 1.0]])).all()

    def test__from_6x6_to_4x4(self):
        array = np.ones((6, 6))
        array[2:4, 2:4] = 2.0

        modified = util.trim_array_2d_around_centre(array_2d=array, new_shape=(4, 4))

        assert (modified == np.array([[1.0, 1.0, 1.0, 1.0],
                                      [1.0, 2.0, 2.0, 1.0],
                                      [1.0, 2.0, 2.0, 1.0],
                                      [1.0, 1.0, 1.0, 1.0]])).all()

    def test__from_5x4_to_3x2(self):
        array = np.ones((5, 4))
        array[2, 1:3] = 2.0

        modified = util.trim_array_2d_around_centre(array_2d=array, new_shape=(3, 2))

        assert (modified == np.array([[1.0, 1.0],
                                      [2.0, 2.0],
                                      [1.0, 1.0]])).all()

    def test__from_4x5_to_2x3(self):
        array = np.ones((4, 5))
        array[1:3, 2] = 2.0

        modified = util.trim_array_2d_around_centre(array_2d=array, new_shape=(2, 3))

        assert (modified == np.array([[1.0, 2.0, 1.0],
                                      [1.0, 2.0, 1.0]])).all()

    def test__x_size_bigger_than_array__raises_error(self):
        array = np.ones((5, 5))
        array[2, 2] = 2.0

        with pytest.raises(ValueError):
            util.trim_array_2d_around_centre(array_2d=array, new_shape=(8, 3))

    def test__y_size_bigger_than_array__raises_error(self):
        array = np.ones((5, 5))
        array[2, 2] = 2.0

        with pytest.raises(ValueError):
            util.trim_array_2d_around_centre(array_2d=array, new_shape=(3, 8))


class TestPad:

    def test__from_even_to_odd_or_odd_to_even__raises_error(self):

        array = np.ones((5, 5))

        with pytest.raises(ValueError):
            util.trim_array_2d_around_centre(array_2d=array, new_shape=(4, 4))

        array = np.ones((6, 6))

        with pytest.raises(ValueError):
            util.trim_array_2d_around_centre(array_2d=array, new_shape=(3, 3))

        array = np.ones((6, 5))

        with pytest.raises(ValueError):
            util.trim_array_2d_around_centre(array_2d=array, new_shape=(3, 3))

        array = np.ones((5, 6))

        with pytest.raises(ValueError):
            util.trim_array_2d_around_centre(array_2d=array, new_shape=(3, 3))

    def test__from_3x3_to_5x5(self):
        array = np.ones((3, 3))
        array[1, 1] = 2.0

        modified = util.pad_array_2d_around_centre(array_2d=array, new_shape=(5, 5))

        assert (modified == np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                      [0.0, 1.0, 1.0, 1.0, 0.0],
                                      [0.0, 1.0, 2.0, 1.0, 0.0],
                                      [0.0, 1.0, 1.0, 1.0, 0.0],
                                      [0.0, 0.0, 0.0, 0.0, 0.0]])).all()

    def test__from_4x4_to_6x6(self):

        array = np.ones((4, 4))
        array[1:3, 1:3] = 2.0

        modified = util.pad_array_2d_around_centre(array_2d=array, new_shape=(6, 6))

        assert (modified == np.array([[0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                                      [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                                      [0.0, 1.0, 2.0, 2.0, 1.0, 0.0],
                                      [0.0, 1.0, 2.0, 2.0, 1.0, 0.0],
                                      [0.0, 1.0, 1.0, 1.0, 1.0, 0.0],
                                      [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],])).all()

    def test__from_3x2_to_5x4(self):
        array = np.ones((3, 2))
        array[1, 0:2] = 2.0

        modified = util.pad_array_2d_around_centre(array_2d=array, new_shape=(5, 4))

        assert (modified == np.array([[0.0, 0.0, 0.0, 0.0],
                                      [0.0, 1.0, 1.0, 0.0],
                                      [0.0, 2.0, 2.0, 0.0],
                                      [0.0, 1.0, 1.0, 0.0],
                                      [0.0, 0.0, 0.0, 0.0]])).all()

    def test__from_2x3_to_4x5(self):
        array = np.ones((2, 3))
        array[0:2, 1] = 2.0

        modified = util.pad_array_2d_around_centre(array_2d=array, new_shape=(4, 5))

        assert (modified == np.array([[0.0, 0.0, 0.0, 0.0, 0.0],
                                      [0.0, 1.0, 2.0, 1.0, 0.0],
                                      [0.0, 1.0, 2.0, 1.0, 0.0],
                                      [0.0, 0.0, 0.0, 0.0, 0.0],])).all()

    def test__x_size_smaller_than_array__raises_error(self):
        array = np.ones((5, 5))
        array[2, 2] = 2.0

        with pytest.raises(ValueError):
            util.trim_array_2d_around_centre(array_2d=array, new_shape=(3, 8))

    def test__y_size_smaller_than_array__raises_error(self):
        array = np.ones((5, 5))
        array[2, 2] = 2.0

        with pytest.raises(ValueError):
            util.trim_array_2d_around_centre(array_2d=array, new_shape=(8, 3))


class TestTrimAroundPixels:

    def test__from_6x6(self):

        array = np.ones((6, 6))
        array[4, 4] = 2.0

        modified = util.trim_array_2d_around_region(array_2d=array, x0=3, x1=6, y0=3, y1=6)

        assert (modified == np.array([[1.0, 1.0, 1.0],
                                      [1.0, 2.0, 1.0],
                                      [1.0, 1.0, 1.0]])).all()

        array = np.ones((6, 6))
        array[0, 0] = 2.0

        modified = util.trim_array_2d_around_region(array_2d=array, x0=0, x1=3, y0=0, y1=3)

        assert (modified == np.array([[2.0, 1.0, 1.0],
                                      [1.0, 1.0, 1.0],
                                      [1.0, 1.0, 1.0]])).all()


        modified = util.trim_array_2d_around_region(array_2d=array, x0=0, x1=2, y0=0, y1=5)

        assert (modified == np.array([[2.0, 1.0],
                                      [1.0, 1.0],
                                      [1.0, 1.0],
                                      [1.0, 1.0],
                                      [1.0, 1.0]])).all()


class TestFits:

    def test__numpy_array_from_fits__3x3_all_ones(self):
        arr = util.numpy_array_from_fits(path=test_data_dir + '3x3_ones.fits', hdu=0)

        assert (arr == np.ones((3, 3))).all()

    def test__numpy_array_from_fits__4x3_all_ones(self):
        arr = util.numpy_array_from_fits(path=test_data_dir + '4x3_ones.fits', hdu=0)

        assert (arr == np.ones((4, 3))).all()

    def test__numpy_array_to_fits__output_and_load(self):
        if os.path.exists(test_data_dir + 'test.fits'):
            os.remove(test_data_dir + 'test.fits')

        arr = np.array([[10., 30., 40.],
                        [92., 19., 20.]])

        util.numpy_array_to_fits(arr, path=test_data_dir + 'test.fits')

        array_load = util.numpy_array_from_fits(path=test_data_dir + 'test.fits', hdu=0)

        assert (arr == array_load).all()


class TestVariancesFromNoise:

    def test__noise_all_1s__variances_all_1s(self):
        noise = np.array([[1.0, 1.0],
                          [1.0, 1.0]])

        assert (util.compute_variances_from_noise(noise) == np.array([[1.0, 1.0],
                                                                      [1.0, 1.0]])).all()

    def test__noise_all_2s__variances_all_4s(self):
        noise = np.array([[2.0, 2.0],
                          [2.0, 2.0]])

        assert (util.compute_variances_from_noise(noise) == np.array([[4.0, 4.0],
                                                                      [4.0, 4.0]])).all()

    def test__noise_all_05s__variances_all_025s(self):
        noise = np.array([[0.5, 0.5],
                          [0.5, 0.5]])

        assert (util.compute_variances_from_noise(noise) == np.array([[0.25, 0.25],
                                                                      [0.25, 0.25]])).all()



import numpy as np
import pytest
from profiling import profiling_data
from profiling import tools


class ImageGridBorder(np.ndarray):

    @property
    def no_pixels(self):
        return self.shape[0]

    def __new__(cls, arr, polynomial_degree=3, centre=(0.0, 0.0), *args, **kwargs):
        border = arr.view(cls)
        border.polynomial_degree = polynomial_degree
        border.centre = centre
        return border

    def grid_to_radii(self, grid):
        """
        Convert coordinates to a circular radius.

        If the coordinates have not been transformed to the profile's geometry, this is performed automatically.

        Parameters
        ----------
        grid

        Returns
        -------
        The radius at those coordinates
        """

        return np.sqrt(np.add(np.square(np.subtract(grid[:, 0], self.centre[0])),
                              np.square(np.subtract(grid[:, 1], self.centre[1]))))

    def grid_to_thetas(self, grid):
        """
        Compute the angle in degrees between the image_grid and plane positive x-axis, defined counter-clockwise.

        Parameters
        ----------
        grid : Union((float, float), ndarray)
            The x and y image_grid of the plane.

        Returns
        ----------
        The angle between the image_grid and the x-axis.
        """
        shifted_grid = np.subtract(grid, self.centre)
        theta_from_x = np.degrees(np.arctan2(shifted_grid[:, 1], shifted_grid[:, 0]))
        theta_from_x[theta_from_x < 0.0] += 360.
        return theta_from_x

    def polynomial_fit_to_border(self, grid):

        border_grid = grid[self]

        return np.polyfit(self.grid_to_thetas(border_grid), self.grid_to_radii(border_grid), self.polynomial_degree)

    def move_factors_from_grid(self, grid):
        """Get the move factor of a coordinate.
         A move-factor defines how far a coordinate outside the source-plane setup_border_pixels must be moved in order
         to lie on it. PlaneCoordinates already within the setup_border_pixels return a move-factor of 1.0, signifying
         they are already within the setup_border_pixels.

        Parameters
        ----------
        grid : ndarray
            The x and y image_grid of the pixel to have its move-factor computed.
        """
        grid_thetas = self.grid_to_thetas(grid)
        grid_radii = self.grid_to_radii(grid)
        poly = self.polynomial_fit_to_border(grid)

        try:
            move_factors = np.polyval(poly, grid_thetas) / grid_radii
        except RuntimeWarning:
            move_factors = 1.0

        move_factors[move_factors > 1.0] = 1.0

        return move_factors

    def relocated_grid_from_grid(self, grid):
        move_factors = self.move_factors_from_grid(grid)
        return np.multiply(grid, move_factors[:, None])


class SubGridBorder(ImageGridBorder):

    @classmethod
    def from_mask(cls, mask, sub_grid_size, polynomial_degree=3, centre=(0.0, 0.0)):
        return cls(mask.border_sub_pixels(sub_grid_size), polynomial_degree, centre)


sub_grid_size = 4

lsst = profiling_data.setup_class(name='LSST', pixel_scale=0.2, sub_grid_size=sub_grid_size)

from profiling.imaging.borders.relocate_grid import original

lsst_border = SubGridBorder.from_mask(mask=lsst.masked_image.mask, sub_grid_size=sub_grid_size)
lsst_border_orig = original.SubGridBorder.from_mask(mask=lsst.masked_image.mask, sub_grid_size=sub_grid_size)

assert (lsst_border_orig.relocated_grid_from_grid(grid=lsst.grids.sub) ==
        pytest.approx(lsst_border.relocated_grid_from_grid(grid=lsst.grids.sub), 1e-4))

euclid = profiling_data.setup_class(name='Euclid', pixel_scale=0.1, sub_grid_size=sub_grid_size)
euclid_border = SubGridBorder.from_mask(mask=euclid.masked_image.mask, sub_grid_size=sub_grid_size)
hst = profiling_data.setup_class(name='HST', pixel_scale=0.05, sub_grid_size=sub_grid_size)
hst_border = SubGridBorder.from_mask(mask=hst.masked_image.mask, sub_grid_size=sub_grid_size)
hst_up = profiling_data.setup_class(name='HSTup', pixel_scale=0.03, sub_grid_size=sub_grid_size)
hst_up_border = SubGridBorder.from_mask(mask=hst_up.masked_image.mask, sub_grid_size=sub_grid_size)
ao = profiling_data.setup_class(name='AO', pixel_scale=0.01, sub_grid_size=sub_grid_size)
ao_border = SubGridBorder.from_mask(mask=ao.masked_image.mask, sub_grid_size=sub_grid_size)


# no jit - no need to run functions first

# euclid_border.relocated_grid_from_grid(euclid.grids.sub)
# hst_border.relocated_grid_from_grid(hst.grids.sub)
# hst_up_border.relocated_grid_from_grid(hst_up.grids.sub)
# ao_border.relocated_grid_from_grid(ao.grids.sub)

@tools.tick_toc_x1
def lsst_solution():
    lsst_border.relocated_grid_from_grid(grid=lsst.grids.sub)


@tools.tick_toc_x1
def euclid_solution():
    euclid_border.relocated_grid_from_grid(grid=euclid.grids.sub)


@tools.tick_toc_x1
def hst_solution():
    hst_border.relocated_grid_from_grid(grid=hst.grids.sub)


@tools.tick_toc_x1
def hst_up_solution():
    hst_up_border.relocated_grid_from_grid(grid=hst_up.grids.sub)


@tools.tick_toc_x1
def ao_solution():
    ao_border.relocated_grid_from_grid(grid=ao.grids.sub)


if __name__ == "__main__":
    lsst_solution()
    euclid_solution()
    hst_solution()
    hst_up_solution()
    ao_solution()



import numpy as np
from analysis import galaxy
from analysis import ray_tracing
from pixelization import pixelization
from profiling import profiling_data
from profiling import tools

from profiles import mass_profiles


class Reconstructor(object):

    def __init__(self, mapping, regularization, image_to_pix, sub_to_pix):
        """The matrices and mappings used to linearly invert and incorrect_fit a data-set.

        Parameters
        -----------
        mapping : ndarray
            The matrix representing the mapping_matrix between reconstructed_image-pixels and data_vector-pixels.
        regularization : ndarray
            The matrix defining how the reconstructed_image's pixels are regularized with one another when fitting the
            data_vector.
        image_to_pix : ndarray
            The mapping_matrix between each masked_image-grid pixel and inversion-grid pixel.
        sub_to_pix : ndarray
            The mapping_matrix between each sub-grid pixel and inversion-grid sub-pixel.
        """
        self.mapping = mapping
        self.mapping_shape = mapping.shape
        self.regularization = regularization
        self.image_to_pix = image_to_pix
        self.sub_to_pix = sub_to_pix

    def covariance_matrix_from_blurred_mapping(self, blurred_mapping, noise_vector):
        """ Compute the curvature_matrix matrix directly - used to integration_old test that our curvature_matrix matrix generator approach
        truly works."""

        covariance_matrix = np.zeros((self.mapping_shape[1], self.mapping_shape[1]))

        for i in range(self.mapping_shape[0]):
            for jx in range(self.mapping_shape[1]):
                for jy in range(self.mapping_shape[1]):
                    covariance_matrix[jx, jy] += blurred_mapping[i, jx] * blurred_mapping[i, jy] \
                                                 / (noise_vector[i] ** 2.0)

        return covariance_matrix


sub_grid_size = 4
psf_size = (41, 41)

sie = mass_profiles.EllipticalIsothermal(centre=(0.010, 0.032), einstein_radius=1.47, axis_ratio=0.849, phi=73.6)
shear = mass_profiles.ExternalShear(magnitude=0.0663, phi=160.5)
lens_galaxy = galaxy.Galaxy(mass_profile_0=sie, mass_profile_1=shear)

source_pix = galaxy.Galaxy(pixelization=pixelization.RectangularRegConst(shape=(19, 19)))

lsst = profiling_data.setup_class(name='LSST', pixel_scale=0.2, sub_grid_size=sub_grid_size, psf_shape=psf_size)
euclid = profiling_data.setup_class(name='Euclid', pixel_scale=0.1, sub_grid_size=sub_grid_size, psf_shape=psf_size)
hst = profiling_data.setup_class(name='HST', pixel_scale=0.05, sub_grid_size=sub_grid_size, psf_shape=psf_size)
hst_up = profiling_data.setup_class(name='HSTup', pixel_scale=0.03, sub_grid_size=sub_grid_size, psf_shape=psf_size)
# ao = profiling_data.setup_class(analysis_path='AO', pixel_scales=0.01, sub_grid_size=sub_grid_size, psf_shape=psf_size)

lsst_tracer = ray_tracing.Tracer(lens_galaxies=[lens_galaxy], source_galaxies=[source_pix],
                                 image_plane_grids=lsst.grids)
euclid_tracer = ray_tracing.Tracer(lens_galaxies=[lens_galaxy], source_galaxies=[source_pix],
                                   image_plane_grids=euclid.grids)
hst_tracer = ray_tracing.Tracer(lens_galaxies=[lens_galaxy], source_galaxies=[source_pix], image_plane_grids=hst.grids)
hst_up_tracer = ray_tracing.Tracer(lens_galaxies=[lens_galaxy], source_galaxies=[source_pix],
                                   image_plane_grids=hst_up.grids)
# ao_tracer = ray_tracing.TracerImagePlane(lens_galaxies=[lens_galaxy], source_galaxies=[source_pix], image_plane_grid=ao.grids)

lsst_recon = lsst_tracer.reconstructors(lsst.borders, cluster_mask=None)
euclid_recon = euclid_tracer.reconstructors(euclid.borders, cluster_mask=None)
hst_recon = hst_tracer.reconstructors(hst.borders, cluster_mask=None)
hst_up_recon = hst_up_tracer.reconstructors(hst_up.borders, cluster_mask=None)
# ao_recon = ao_tracer.reconstructors_from_source_plane(ao.borders, cluster_mask=None)

lsst_recon = Reconstructor(lsst_recon.mapping, lsst_recon.regularization, lsst_recon.image_to_pix,
                           lsst_recon.sub_to_pix)
euclid_recon = Reconstructor(euclid_recon.mapping, euclid_recon.regularization, euclid_recon.image_to_pix,
                             euclid_recon.sub_to_pix)
hst_recon = Reconstructor(hst_recon.mapping, hst_recon.regularization, hst_recon.image_to_pix,
                          hst_recon.sub_to_pix)
hst_up_recon = Reconstructor(hst_up_recon.mapping, hst_up_recon.regularization, hst_up_recon.image_to_pix,
                             hst_up_recon.sub_to_pix)
# ao_recon = Reconstructor(ao_recon.mapping_matrix, ao_recon.regularization_matrix, ao_recon.image_to_pixelization,
#                            ao_recon.sub_to_pixelization)

lsst_blurred_mapping = lsst.masked_image.convolver_mapping_matrix.convolve_mapping_matrix(lsst_recon.mapping)
euclid_blurred_mapping = euclid.masked_image.convolver_mapping_matrix.convolve_mapping_matrix(euclid_recon.mapping)
hst_blurred_mapping = hst.masked_image.convolver_mapping_matrix.convolve_mapping_matrix(hst_recon.mapping)
hst_up_blurred_mapping = hst_up.masked_image.convolver_mapping_matrix.convolve_mapping_matrix(hst_up_recon.mapping)


# ao_blurred_mapping = ao.masked_image.convolver_mapping_matrix.convolve_mapping_matrix_jit(ao_recon.mapping_matrix)

@tools.tick_toc_x1
def lsst_solution():
    lsst_recon.covariance_matrix_from_blurred_mapping(lsst_blurred_mapping, lsst.masked_image.estimated_noise_map)


@tools.tick_toc_x1
def euclid_solution():
    euclid_recon.covariance_matrix_from_blurred_mapping(euclid_blurred_mapping, euclid.masked_image.estimated_noise_map)


@tools.tick_toc_x1
def hst_solution():
    hst_recon.covariance_matrix_from_blurred_mapping(hst_blurred_mapping, hst.masked_image.estimated_noise_map)


@tools.tick_toc_x1
def hst_up_solution():
    hst_up_recon.covariance_matrix_from_blurred_mapping(hst_up_blurred_mapping, hst_up.masked_image.estimated_noise_map)


@tools.tick_toc_x1
def ao_solution():
    ao_recon.curvature_matrix_from_blurred_mapping_matrix(ao_blurred_mapping, ao.masked_image.estimated_noise_map)


if __name__ == "__main__":
    lsst_solution()
    euclid_solution()
    hst_solution()
    hst_up_solution()
    ao_solution()



"""Reset Config."""

from autolens import conf
from .base import Base, current_directory

config_path = "{}/config".format(current_directory)


class ResetConfig(Base):
    """Reset Config!"""

    def run(self):
        if conf.is_config(config_path):
            if input("Are you sure? This will reset the state of your config. (y/n)\n").lower() != 'y':
                return
            conf.remove_config(config_path)
        conf.copy_default(config_path)



from itertools import count

import numpy as np

from autolens import exc
from autolens.profiles import light_profiles as lp
from autolens.profiles import mass_profiles as mp


def is_light_profile(obj):
    return isinstance(obj, lp.LightProfile)


def is_mass_profile(obj):
    return isinstance(obj, mp.MassProfile)


class Galaxy(object):
    """
    @DynamicAttrs
    """

    def __init__(self, redshift=None, pixelization=None, regularization=None, hyper_galaxy=None, **kwargs):
        """
        Represents a real galaxy. This could be a lens galaxy or source galaxy. Note that a lens galaxy must have mass
        profiles

        Parameters
        ----------
        redshift: float
            The redshift of this galaxy
        light_profiles: [LightProfile]
            A list of light profiles describing the light profiles of this galaxy
        mass_profiles: [MassProfile]
            A list of mass profiles describing the mass profiles of this galaxy
        """
        self.redshift = redshift

        for name, val in kwargs.items():
            setattr(self, name, val)

        self.pixelization = pixelization
        self.regularization = regularization

        if self.has_pixelization and not self.has_regularization:
            raise exc.GalaxyException('If the galaxy has a pixelization, it must also have a regularization.')
        if not self.has_pixelization and self.has_regularization:
            raise exc.GalaxyException('If the galaxy has a regularization, it must also have a pixelization.')

        self.hyper_galaxy = hyper_galaxy

    @property
    def light_profiles(self):
        return [value for value in self.__dict__.values() if is_light_profile(value)]

    @property
    def mass_profiles(self):
        return [value for value in self.__dict__.values() if is_mass_profile(value)]

    @property
    def has_redshift(self):
        return self.redshift is not None

    @property
    def has_pixelization(self):
        """
        Returns
        -------
        True iff this galaxy has an associated inversion
        """
        return self.pixelization is not None

    @property
    def has_regularization(self):
        """
        Returns
        -------
        True iff this galaxy has an associated inversion
        """
        return self.regularization is not None

    @property
    def has_hyper_galaxy(self):
        """
        Returns
        -------
        True iff this galaxy has an associated hyper galaxy
        """
        return self.hyper_galaxy is not None

    @property
    def has_light_profile(self):
        """
        Returns
        -------
        True iff there is one or more mass or light profiles associated with this galaxy
        """
        return len(self.light_profiles) > 0

    @property
    def has_mass_profile(self):
        """
        Returns
        -------
        True iff there is one or more mass or light profiles associated with this galaxy
        """
        return len(self.mass_profiles) > 0

    @property
    def has_profile(self):
        """
        Returns
        -------
        True iff there is one or more mass or light profiles associated with this galaxy
        """
        return len(self.mass_profiles) + len(self.light_profiles) > 0

    def __repr__(self):
        string = "Redshift: {}".format(self.redshift)
        if self.pixelization:
            string += str(self.pixelization)
        if self.hyper_galaxy:
            string += str(self.hyper_galaxy)
        if self.light_profiles:
            string += "\nLight Profiles:\n{}".format("\n".join(map(str, self.light_profiles)))
        if self.mass_profiles:
            string += "\nMass Profiles:\n{}".format("\n".join(map(str, self.mass_profiles)))
        return string

    def intensities_from_grid(self, grid):
        if self.light_profiles is not None and len(self.light_profiles) > 0:
            return sum(map(lambda p: p.intensities_from_grid(grid), self.light_profiles))
        else:
            return np.zeros((grid.shape[0],))

    def luminosity_within_circle(self, radius):
        """
        Compute the total luminosity of the galaxy's light profiles within a circle of specified radius.

        See *light_profiles.luminosity_within_circle* for details of how this is performed.

        Parameters
        ----------
        radius : float
            The radius of the circle to compute the luminosity within.

        Returns
        -------
        luminosity : float
            The total combined luminosity within the specified circle.
        """
        return sum(map(lambda p: p.luminosity_within_circle(radius), self.light_profiles))

    def luminosity_within_ellipse(self, major_axis):
        """
        Compute the total luminosity of the galaxy's light profiles, within an ellipse of specified major axis. This 
        is performed via integration_old of each light profile and is centred, oriented and  aligned with each light
        model_mapper's individual geometry.

        See *light_profiles.luminosity_within_ellipse* for details of how this is performed.

        Parameters
        ----------
        major_axis: float
            The major-axis of the ellipse to compute the luminosity within.
        Returns
        -------
        intensity : float
            The total luminosity within the specified ellipse.
        """
        return sum(map(lambda p: p.luminosity_within_ellipse(major_axis), self.light_profiles))

    def luminosity_within_ellipse_individual(self, major_axis):
        """
        Compute the individual total luminosity of each light profile in the galaxy, within an ellipse of 
        specified major axis.

        See *light_profiles.luminosity_within_ellipse* for details of how this is performed.

        Parameters
        ----------
        major_axis: float
            The major-axis of the ellipse to compute the luminosity within.
        Returns
        -------
        intensity : [float]
            The total luminosity within the specified ellipse.
        """
        return list(map(lambda p: p.luminosity_within_ellipse(major_axis), self.light_profiles))

    def surface_density_from_grid(self, grid):
        """

        Compute the summed surface density of the galaxy's mass profiles at a given set of image_grid.

        See *mass_profiles* module for details of how this is performed.

        Parameters
        ----------
        grid : ndarray
            The x and y image_grid of the image_grid

        Returns
        ----------
        The summed values of surface density at the given image_grid.
        """
        if self.mass_profiles is not None and len(self.mass_profiles) > 0:
            return sum(map(lambda p: p.surface_density_from_grid(grid), self.mass_profiles))
        else:
            return np.zeros((grid.shape[0],))

    def potential_from_grid(self, grid):
        """
        Compute the summed gravitational potential of the galaxy's mass profiles at a given set of image_grid.

        See *mass_profiles* module for details of how this is performed.

        Parameters
        ----------
        grid : ndarray
            The x and y image_grid of the image_grid

        Returns
        ----------
        The summed values of gravitational potential at the given image_grid.
        """
        if self.mass_profiles is not None and len(self.mass_profiles) > 0:
            return sum(map(lambda p: p.potential_from_grid(grid), self.mass_profiles))
        else:
            return np.zeros((grid.shape[0],))

    def deflections_from_grid(self, grid):
        if self.mass_profiles is not None and len(self.mass_profiles) > 0:
            return sum(map(lambda p: p.deflections_from_grid(grid), self.mass_profiles))
        else:
            return np.full((grid.shape[0], 2), 0.0)

    def dimensionless_mass_within_circles(self, radius):
        """
        Compute the total dimensionless mass of the galaxy's mass profiles within a circle of specified radius.

        See *mass_profiles.dimensionless_mass_within_circles* for details of how this is performed.


        Parameters
        ----------
        radius : float
            The radius of the circle to compute the dimensionless mass within.

        Returns
        -------
        dimensionless_mass : float
            The total dimensionless mass within the specified circle.
        """
        return sum(map(lambda p: p.dimensionless_mass_within_circle(radius), self.mass_profiles))

    def dimensionless_mass_within_ellipses(self, major_axis):
        """
        Compute the total dimensionless mass of the galaxy's mass profiles within an ellipse of specified major_axis.

        See *mass_profiles.dimensionless_mass_within_ellipses* for details of how this is performed.


        Parameters
        ----------
        major_axis : float
            The major axis of the ellipse

        Returns
        -------
        dimensionless_mass : float
            The total dimensionless mass within the specified circle.
        """
        return sum(map(lambda p: p.dimensionless_mass_within_ellipse(major_axis), self.mass_profiles))


# TODO : Should galaxy masked_image and minimum value be in the constructor (they aren't free parameters)?

class HyperGalaxy(object):
    _ids = count()

    def __init__(self, contribution_factor=0.0, noise_factor=0.0, noise_power=1.0):
        """Class for scaling the noise_map in the different galaxies of an masked_image (e.g. the lens, source).

        Parameters
        -----------
        contribution_factor : float
            Factor that adjusts how much of the galaxy's light is attributed to the contribution mappers.
        noise_factor : float
            Factor by which the noise_map is increased in the regions of the galaxy's contribution mappers.
        noise_power : float
            The power to which the contribution mappers is raised when scaling the noise_map.
        """
        self.contribution_factor = contribution_factor
        self.noise_factor = noise_factor
        self.noise_power = noise_power

        self.component_number = next(self._ids)

    def contributions_from_hyper_images(self, hyper_model_image, hyper_galaxy_image, minimum_value):
        """Compute the contribution mappers of a galaxy, which represents the fraction of flux in each pixel that \
        galaxy can be attributed to contain.

        This is computed by dividing that galaxy's flux by the total flux in that pixel, and then scaling by the \
        maximum flux such that the contribution mappers ranges between 0 and 1.

        Parameters
        -----------
        hyper_model_image : ndarray
            The model masked_image of the observed data_vector (from a previous lensing phase). This tells us the total
            light attributed to each masked_image pixel by the model.
        hyper_galaxy_image : ndarray
            A model masked_image of the galaxy (e.g the lens light profile or source reconstructed_image) computed from
            a previous lensing.
        minimum_value : float
            The minimum fractional flux a pixel must contain to not be rounded to 0.
        """
        contributions = np.divide(hyper_galaxy_image, np.add(hyper_model_image, self.contribution_factor))
        contributions = np.divide(contributions, np.max(contributions))
        contributions[contributions < minimum_value] = 0.0
        return contributions

    def scaled_noise_from_contributions(self, noise, contributions):
        """Compute a scaled galaxy noise_map mappers from a baseline nosie mappers.

        This uses the galaxy contribution mappers with their noise_map scaling hyper-parameters.

        Parameters
        -----------
        noise : ndarray
            The noise_map before scaling (this may already have the background scaled in HyperImage)
        contributions : ndarray
            The galaxy contribution mappers.
        """
        return self.noise_factor * (noise * contributions) ** self.noise_power

    def __eq__(self, other):
        if isinstance(other, HyperGalaxy):
            return self.contribution_factor == other.contribution_factor and \
                   self.noise_factor == other.noise_factor and \
                   self.noise_power == other.noise_power
        return False

    def __str__(self):
        return "\n".join(["{}: {}".format(k, v) for k, v in self.__dict__.items()])


class Redshift(object):
    def __init__(self, redshift):
        self.redshift = redshift

    def __str__(self):
        return str(self.redshift)



import numba
import numpy as np
import pytest
from profiling import profiling_data
from profiling import tools

from profiles import geometry_profiles


class EllipticalIsothermal(geometry_profiles.EllipticalProfile):

    def __init__(self, centre=(0.0, 0.0), axis_ratio=0.9, phi=0.0, einstein_radius=1.0):
        """
        Represents an elliptical isothermal density distribution, which is equivalent to the elliptical power-law
        density distribution for the value slope=2.0

        Parameters
        ----------
        centre: (float, float)
            The image_grid of the centre of the profiles
        axis_ratio : float
            Elliptical mass profile's minor-to-major axis ratio (b/a)
        phi : float
            Rotation angle of mass profile's ellipse counter-clockwise from positive x-axis
        einstein_radius : float
            Einstein radius of power-law mass profiles
        """

        super(EllipticalIsothermal, self).__init__(centre, axis_ratio, phi)
        self.einstein_radius = einstein_radius
        self.slope = 2.0

    @property
    def einstein_radius_rescaled(self):
        """Rescale the einstein radius by slope and axis_ratio, to reduce its degeneracy with other mass-profiles
        parameters"""
        return ((3 - self.slope) / (1 + self.axis_ratio)) * self.einstein_radius ** (self.slope - 1)

    def deflections_from_grid(self, grid):
        """
        Calculate the deflection angles at a given set of gridded coordinates.

        Parameters
        ----------
        grid : mask.ImageGrid
            The grid of coordinates the deflection angles are computed on.
        """

        try:
            factor = 2.0 * self.einstein_radius_rescaled * self.axis_ratio / np.sqrt(1 - self.axis_ratio ** 2)

            psi = np.sqrt(np.add(np.multiply(self.axis_ratio ** 2, np.square(grid[:, 0])), np.square(grid[:, 1])))

            deflection_x = np.arctan(np.divide(np.multiply(np.sqrt(1 - self.axis_ratio ** 2), grid[:, 0]), psi))
            deflection_y = np.arctanh(np.divide(np.multiply(np.sqrt(1 - self.axis_ratio ** 2), grid[:, 1]), psi))

            return np.multiply(factor, np.vstack((deflection_x, deflection_y)).T)
        except ZeroDivisionError:
            return self.grid_radius_to_cartesian(grid, np.full(grid.shape[0], 2.0 * self.einstein_radius_rescaled))

    def deflections_from_grid_jitted(self, grid):
        return self.deflections_from_grid_jit(grid, self.axis_ratio, self.einstein_radius_rescaled)

    @staticmethod
    @numba.jit(nopython=True)
    def deflections_from_grid_jit(grid, axis_ratio, einstein_radius_rescaled):

        factor = 2.0 * einstein_radius_rescaled * axis_ratio / np.sqrt(1 - axis_ratio ** 2)
        q1 = np.sqrt(1 - axis_ratio ** 2)

        deflections = np.zeros(grid.shape)
        for i in range(deflections.shape[0]):
            psi = np.multiply(np.divide(1.0, q1),
                              np.sqrt(np.add(np.multiply(np.square(axis_ratio), np.square(grid[i, 0])),
                                             np.square(grid[i, 1]))))

            deflections[i, 0] = np.multiply(factor, np.arctan(np.divide(grid[i, 0], psi)))
            deflections[i, 1] = np.multiply(factor, np.arctanh(np.divide(grid[i, 1], psi)))

        return deflections


sie = EllipticalIsothermal(centre=(0.0, 0.0), axis_ratio=0.8, phi=90.0, einstein_radius=1.4)

subgrd_size = 4

lsst = profiling_data.setup_class(name='LSST', pixel_scale=0.2, sub_grid_size=subgrd_size)
euclid = profiling_data.setup_class(name='Euclid', pixel_scale=0.1, sub_grid_size=subgrd_size)
hst = profiling_data.setup_class(name='HST', pixel_scale=0.05, sub_grid_size=subgrd_size)
hst_up = profiling_data.setup_class(name='HSTup', pixel_scale=0.03, sub_grid_size=subgrd_size)
ao = profiling_data.setup_class(name='AO', pixel_scale=0.01, sub_grid_size=subgrd_size)

assert (sie.deflections_from_grid(grid=lsst.coords.sub_grid_coords) ==
        pytest.approx(sie.deflections_from_grid_jitted(grid=lsst.coords.sub_grid_coords), 1e-4))


@tools.tick_toc_x10
def lsst_solution():
    sie.deflections_from_grid_jitted(grid=lsst.coords.sub_grid_coords)


@tools.tick_toc_x10
def euclid_solution():
    sie.deflections_from_grid_jitted(grid=euclid.coords.sub_grid_coords)


@tools.tick_toc_x10
def hst_solution():
    sie.deflections_from_grid_jitted(grid=hst.coords.sub_grid_coords)


@tools.tick_toc_x10
def hst_up_solution():
    sie.deflections_from_grid_jitted(grid=hst_up.coords.sub_grid_coords)


@tools.tick_toc_x10
def ao_solution():
    sie.deflections_from_grid_jitted(grid=ao.coords.sub_grid_coords)


if __name__ == "__main__":
    lsst_solution()
    euclid_solution()
    hst_solution()
    hst_up_solution()
    ao_solution()



from runners.lens_and_source import initializer
from autolens.imaging import scaled_array
from autolens.imaging import image
from autolens import conf
import os

dirpath = os.path.dirname(os.path.realpath(__file__))
conf.instance.output_path = os.path.expanduser("~/")


def load_image(data_name, pixel_scale, image_hdu, noise_hdu, psf_hdu, psf_trimmed_shape=None,
               effective_exposure_time=None):
    data_dir = "{}/../data/{}".format(dirpath, data_name)

    data = scaled_array.ScaledSquarePixelArray.from_fits_with_pixel_scale(file_path=data_dir, hdu=image_hdu, pixel_scale=pixel_scale)
    data = data.trim_around_centre((301, 301))
    background_noise = scaled_array.ScaledSquarePixelArray.from_fits_with_pixel_scale(file_path=data_dir, hdu=noise_hdu)
    background_noise = background_noise.trim_around_centre((301, 301))
    psf = image.PSF.from_fits_with_scale(file_path=data_dir, hdu=psf_hdu)
    if psf_trimmed_shape is not None:
        psf = psf.trim_around_centre(psf_trimmed_shape)

    if isinstance(effective_exposure_time, float):
        effective_exposure_time = scaled_array.ScaledSquarePixelArray.single_value(value=effective_exposure_time, shape=data.shape,
                                                                                   pixel_scale=pixel_scale)

    return image.PreparatoryImage(array=data, pixel_scales=pixel_scale, psf=psf, background_noise_map=background_noise,
                                  effective_exposure_map=effective_exposure_time)


im = load_image(data_name='slacs05_bg/slacs_4_post.fits', pixel_scale=0.05, image_hdu=1, noise_hdu=2, psf_hdu=3,
                psf_trimmed_shape=(41, 41), effective_exposure_time=288.0)

im.background_noise_map = 1.0 / im.background_noise_map
im.noise_map = im.estimated_noise_map
pipeline = initializer.make()
pipeline.run(im)



"""
Initialize the lens model of a 2-plane lens+source system.

This pipeline is composed of 3 phases:

1) Fit and subtract the lens light using an elliptical Sersic.
2) Fit the source using the lens subtracted image from phase 1, with an SIE mass model (lens) and elliptical
Sersic (source).
3) Fit the lens and source simulataneously, using priors based on the results of phases 1 and 2.
"""

pipeline_name = 'initialize_all'


def make():
    from autolens.pipeline import phase
    from autolens.pipeline import pipeline
    from autolens.autofit import non_linear as nl
    from autolens.galaxy import galaxy_model as gp
    from autolens.profiles import light_profiles as lp
    from autolens.profiles import mass_profiles as mp

    phase1 = phase.LensSourcePlanePhase(lens_galaxies=[gp.GalaxyModel(light=lp.EllipticalSersic,
                                                                      mass=mp.EllipticalIsothermal)],
                                        source_galaxies=[gp.GalaxyModel(light=lp.EllipticalSersic)],
                                        optimizer_class=nl.MultiNest, phase_name='ph1_fit_all')

    phase1.optimizer.n_live_points = 80
    phase1.optimizer.sampling_efficiency = 0.6

    return pipeline.PipelineImaging(pipeline_name, phase1)



"""
Initialize the lens model of a 2-plane lens+source system.

This pipeline is composed of 3 phases:

1) Fit and subtract the lens light using an elliptical Sersic.
2) Fit the source using the lens subtracted image from phase 1, with an SIE mass model (lens) and elliptical
Sersic (source).
3) Fit the lens and source simultaneously, using priors based on the results of phases 1 and 2.
"""

pipeline_name = 'initializer'


def make():
    from autolens.pipeline import phase
    from autolens.pipeline import pipeline
    from autolens.autofit import non_linear as nl
    from autolens.imaging import mask
    from autolens.galaxy import galaxy_model as gp
    from autolens.profiles import light_profiles as lp
    from autolens.profiles import mass_profiles as mp

    phase1 = phase.LensPlanePhase(lens_galaxies=dict(lens_galaxy=gp.GalaxyModel(sersic=lp.EllipticalSersic)),
                                  optimizer_class=nl.MultiNest, phase_name='ph1_subtract_lens')

    phase1.optimizer.n_live_points = 50
    phase1.optimizer.sampling_efficiency = 0.8

    class LensSubtractedPhase(phase.LensSourcePlanePhase):

        def modify_image(self, masked_image, previous_results):
            return previous_results.last.lens_subtracted_image

        def pass_priors(self, previous_results):
            self.lens_galaxies.lens_galaxy.sie.centre_0 = previous_results.last. \
                variable.lens_galaxies.lens_galaxy.sersic.centre_0
            self.lens_galaxies.lens_galaxy.sie.centre_1 = previous_results.last. \
                variable.lens_galaxies.lens_galaxy.sersic.centre_1

    def annular_mask_function(img):
        return mask.Mask.annular(img.shape, pixel_scale=img.pixel_scale, inner_radius_arcsec=0.4,
                                 outer_radius_arcsec=3.)

    phase2 = LensSubtractedPhase(lens_galaxies=dict(lens_galaxy=gp.GalaxyModel(sie=mp.SphericalIsothermal)),
                                 source_galaxies=dict(source_galaxy=gp.GalaxyModel(sersic=lp.EllipticalSersic)),
                                 optimizer_class=nl.MultiNest, mask_function=annular_mask_function,
                                 phase_name='ph2_fit_source')

    phase2.optimizer.n_live_points = 60
    phase2.optimizer.sampling_efficiency = 0.6

    class LensSourcePhase(phase.LensSourcePlanePhase):

        def pass_priors(self, previous_results):
            self.lens_galaxies.lens_galaxy = gp.GalaxyModel(
                sersic=previous_results.first.variable.lens_galaxies.lens_galaxy.sersic,
                sie=previous_results.last.variable.lens_galaxies.lens_galaxy.sie)
            self.source_galaxies = previous_results.last.variable.source_galaxies

    phase3 = LensSourcePhase(optimizer_class=nl.MultiNest, phase_name='ph3_fit_all')

    phase3.optimizer.n_live_points = 60
    phase3.optimizer.sampling_efficiency = 0.8

    return pipeline.PipelineImaging(pipeline_name, phase1, phase2, phase3)



import numpy as np
import scipy.signal
from scipy.stats import norm

from autolens import exc
from autolens.imaging import imaging_util
from autolens.imaging.scaled_array import ScaledSquarePixelArray, Array

import logging


logger = logging.getLogger(__name__)


class Image(ScaledSquarePixelArray):

    def __init__(self, array, pixel_scale, psf, noise_map, background_noise_map=None):
        """
        A 2d array representing a real or simulated _data.

        Parameters
        ----------
        array : ndarray
            An array of the _data.
        pixel_scale : float
            The scale of each pixel in arc seconds
        psf : PSF
            An array describing the PSF of the _data.
        noise_map : Array
            An array describing the total noise_map in each _data pixel.
        background_noise_map : Array
            An array describing the background noise_map in each _data pixel (used for hyper_image background noise_map
            scaling).
        """
        super(Image, self).__init__(array, pixel_scale)
        self.psf = psf
        self.noise_map = noise_map
        self.background_noise_map = background_noise_map

    def __array_finalize__(self, obj):
        super(Image, self).__array_finalize__(obj)
        if isinstance(obj, Image):
            try:
                self.psf = obj.psf
                self.noise_map = obj.noise_map
                self.background_noise_map = obj.background_noise_map
            except AttributeError:
                logger.debug("Original object in Image.__array_finalize__ missing one or more attributes")

    def trim_image_and_noise_around_centre(self, new_shape):

        if self.background_noise_map is not None:
            background_noise_map = self.background_noise_map.trim_around_centre(new_shape)
        else:
            background_noise_map = None

        return Image(array=self.trim_around_centre(new_shape), pixel_scale=self.pixel_scale, psf=self.psf,
                     noise_map=self.noise_map.trim_around_centre(new_shape), background_noise_map=background_noise_map)

    def trim_image_and_noise_around_region(self, x0, x1, y0, y1):

        if self.background_noise_map is not None:
            background_noise_map = self.background_noise_map.trim_around_region(x0, x1, y0, y1)
        else:
            background_noise_map = None

        return Image(array=self.trim_around_region(x0, x1, y0, y1), pixel_scale=self.pixel_scale, psf=self.psf,
                     noise_map=self.noise_map.trim_around_region(x0, x1, y0, y1),
                     background_noise_map=background_noise_map)

    @property
    def signal_to_noise_map(self):
        """The estimated signal-to-noise_map mappers of the _data."""
        signal_to_noise_map = np.divide(self, self.noise_map)
        signal_to_noise_map[signal_to_noise_map < 0] = 0
        return signal_to_noise_map

    @property
    def signal_to_noise_max(self):
        """The maximum value of signal-to-noise_map in an _data pixel in the _data's signal-to-noise_map mappers"""
        return np.max(self.signal_to_noise_map)


class PreparatoryImage(Image):

    def __init__(self, array, pixel_scale, psf, noise_map=None, background_noise_map=None, poisson_noise_map=None,
                 exposure_time=None, effective_exposure_map=None):
        """
        A 2d array representing an _data, including preparatory components which are not needed for the actual lens
        analysis but help set up the noise_map, background sky, etc.

        Parameters
        ----------
        array : ndarray
            An array of the _data.
        pixel_scale : float
            The scale of each pixel in arc seconds
        psf : PSF
            An array describing the PSF of the _data.
        noise_map : Array
            An array describing the total noise_map in each _data pixel.
        background_noise_map : Array
            An array describing the background noise_map in each _data pixel (used for hyper_image background noise_map
            scaling).
        poisson_noise_map : ndarray
            An array describing the poisson noise_map in each _data pixel (used for checking the _data units are
            sensible).
        exposure_time : float
            The overall exposure time of the _data.
        effective_exposure_map : ndarray
            An array representing the effective exposure time of each pixel.
        """
        super(PreparatoryImage, self).__init__(array, pixel_scale, psf, noise_map, background_noise_map)
        self.poisson_noise_map = poisson_noise_map
        self.exposure_time = exposure_time
        self.effective_exposure_map = effective_exposure_map

    @classmethod
    def simulate(cls, array, pixel_scale, exposure_time, psf=None, background_sky_level=None,
                 add_noise=False, seed=-1):

        effective_exposure_map = ScaledSquarePixelArray.single_value(value=exposure_time, shape=array.shape,
                                                                     pixel_scale=pixel_scale)
        if background_sky_level is not None:
            background_sky_map = ScaledSquarePixelArray.single_value(value=background_sky_level, shape=array.shape,
                                                                     pixel_scale=pixel_scale)
        else:
            background_sky_map = None

        return cls.simulate_variable_arrays(array=array, pixel_scale=pixel_scale,
                                            effective_exposure_map=effective_exposure_map, psf=psf,
                                            background_sky_map=background_sky_map,
                                            add_noise=add_noise, seed=seed)

    @classmethod
    def simulate_variable_arrays(cls, array, pixel_scale, effective_exposure_map, psf=None, background_sky_map=None,
                                 add_noise=True, seed=-1):
        """
        Create a realistic simulated _data by applying effects to a plain simulated _data.

        Parameters
        ----------
        array: ndarray
            The _data before simulating (e.g. the lens and source galaxies before optics blurring and CCD read-out).
        pixel_scale: float
            The scale of each pixel in arc seconds
        effective_exposure_map : ndarray
            An array representing the effective exposure time of each pixel.
        psf: PSF
            An array describing the PSF the simulated _data is blurred with.
        background_sky_map : ndarray
            The value of background sky in every _data pixel (electrons per second).
        add_noise: Bool
            If True poisson noise_map is simulated and added to the _data, based on the total counts in each _data
            pixel
        seed: int
            A seed for random noise_map generation
        """

        if background_sky_map is not None:
            array += background_sky_map

        if psf is not None:
            array = psf.convolve(array)
            array = cls.trim_psf_edges(array, psf)
            effective_exposure_map = cls.trim_psf_edges(effective_exposure_map, psf)
            if background_sky_map is not None:
                background_sky_map = cls.trim_psf_edges(background_sky_map, psf)

        if add_noise is True:
            array += generate_poisson_noise(array, effective_exposure_map, seed)
            array_counts = np.multiply(array, effective_exposure_map)
            noise_map = np.divide(np.sqrt(array_counts), effective_exposure_map)
        else:
            noise_map = None

        if background_sky_map is not None:
            array -= background_sky_map

        # ESTIMATE THE BACKGROUND NOISE MAP FROM THE IMAGE

        if background_sky_map is not None:
            background_noise_map_counts = np.sqrt(np.multiply(background_sky_map, effective_exposure_map))
            background_noise_map = np.divide(background_noise_map_counts, effective_exposure_map)
        else:
            background_noise_map = None

        # ESTIMATE HTE POISSON NOISE MAP FROM THE IMAGE

        array_counts = np.multiply(array, effective_exposure_map)
        poisson_noise_map = np.divide(np.sqrt(np.abs(array_counts)), effective_exposure_map)

        return PreparatoryImage(array, pixel_scale=pixel_scale, noise_map=noise_map, psf=psf,
                                background_noise_map=background_noise_map, poisson_noise_map=poisson_noise_map,
                                effective_exposure_map=effective_exposure_map)

    @classmethod
    def simulate_to_target_signal_to_noise(cls, array, pixel_scale, target_signal_to_noise, effective_exposure_map,
                                           psf=None, background_sky_map=None, seed=-1):
        """
        Create a realistic simulated _data by applying effects to a plain simulated _data.

        Parameters
        ----------
        target_signal_to_noise
        array: ndarray
            The _data before simulating (e.g. the lens and source galaxies before optics blurring and CCD read-out).
        pixel_scale: float
            The scale of each pixel in arc seconds
        effective_exposure_map : ndarray
            An array representing the effective exposure time of each pixel.
        psf: PSF
            An array describing the PSF the simulated _data is blurred with.
        background_sky_map : ndarray
            The value of background sky in every _data pixel (electrons per second).
        seed: int
            A seed for random noise_map generation
        """

        max_index = np.unravel_index(array.argmax(), array.shape)
        max_array = array[max_index]
        max_effective_exposure_time = effective_exposure_map[max_index]
        max_array_counts = np.multiply(max_array, max_effective_exposure_time)
        if background_sky_map is not None:
            max_background_sky_map = background_sky_map[max_index]
            max_background_sky_map_counts = np.multiply(max_background_sky_map, max_effective_exposure_time)
        else:
            max_background_sky_map_counts = None

        scale_factor = 1.

        if background_sky_map is None:
            scale_factor = target_signal_to_noise ** 2.0 / max_array_counts
        elif background_sky_map is not None:
            scale_factor = (max_array_counts + max_background_sky_map_counts) * target_signal_to_noise ** 2.0 \
                           / max_array_counts ** 2.0

        scaled_effective_exposure_time = np.multiply(scale_factor, effective_exposure_map)

        return cls.simulate_variable_arrays(array=array, pixel_scale=pixel_scale,
                                            effective_exposure_map=scaled_effective_exposure_time,
                                            psf=psf, background_sky_map=background_sky_map,
                                            add_noise=True, seed=seed)

    def __array_finalize__(self, obj):
        super(PreparatoryImage, self).__array_finalize__(obj)
        if isinstance(obj, PreparatoryImage):
            self.psf = obj.psf
            self.noise_map = obj.noise_map
            self.background_noise_map = obj.background_noise_map
            self.poisson_noise_map = obj.poisson_noise_map
            self.effective_exposure_map = obj.effective_exposure_map

    @staticmethod
    def trim_psf_edges(array, psf):
        if psf is not None:
            psf_cut_x = np.int(np.ceil(psf.shape[0] / 2)) - 1
            psf_cut_y = np.int(np.ceil(psf.shape[1] / 2)) - 1
            array_x = np.int(array.shape[0])
            array_y = np.int(array.shape[1])
            return array[psf_cut_x:array_x - psf_cut_x, psf_cut_y:array_y - psf_cut_y]
        else:
            return array

    def electrons_per_second_to_counts(self, array):
        """
        For an array (in electrons per second) and an exposure time mappers, return an array in units counts.

        Parameters
        ----------
        array : ndarray
            The array the values are to be converted from electrons per seconds to counts.
        """
        return np.multiply(array, self.effective_exposure_map)

    def counts_to_electrons_per_second(self, array):
        """
        For an array (in counts) and an exposure time mappers, convert the array to units electrons per second

        Parameters
        ----------
        array : ndarray
            The array the values are to be converted from counts to electrons per second.
        """
        return np.divide(array, self.effective_exposure_map)

    @property
    def image_counts(self):
        """The _data in units of counts."""
        return self.electrons_per_second_to_counts(self)

    @property
    def background_noise_map_counts(self):
        """ The background noise_map mappers in units of counts."""
        return self.electrons_per_second_to_counts(self.background_noise_map)

    @property
    def estimated_noise_map_counts(self):
        """ The estimated noise_map mappers of the _data (using its background noise_map mappers and _data values
        in counts) in counts.
        """
        return np.sqrt((np.abs(self.image_counts) + np.square(self.background_noise_map_counts)))

    @property
    def estimated_noise_map(self):
        """ The estimated noise_map mappers of the _data (using its background noise_map mappers and _data values
        in counts) in electrons per second.
        """
        return self.counts_to_electrons_per_second(self.estimated_noise_map_counts)

    def background_noise_from_edges(self, no_edges):
        """Estimate the background signal_to_noise_ratio by binning data_to_image located at the edge(s) of an _data
        into a histogram and fitting a Gaussian profiles to this histogram. The standard deviation (sigma) of this
        Gaussian gives a signal_to_noise_ratio estimate.

        Parameters
        ----------
        no_edges : int
            Number of edges used to estimate the background signal_to_noise_ratio.

        """

        edges = []

        for edge_no in range(no_edges):
            top_edge = self[edge_no, edge_no:self.shape[1] - edge_no]
            bottom_edge = self[self.shape[0] - 1 - edge_no, edge_no:self.shape[1] - edge_no]
            left_edge = self[edge_no + 1:self.shape[0] - 1 - edge_no, edge_no]
            right_edge = self[edge_no + 1:self.shape[0] - 1 - edge_no, self.shape[1] - 1 - edge_no]

            edges = np.concatenate((edges, top_edge, bottom_edge, right_edge, left_edge))

        return norm.fit(edges)[1]


class NoiseMap(ScaledSquarePixelArray):

    @classmethod
    def from_weight_map(cls, pixel_scale, weight_map):
        np.seterr(divide='ignore')
        noise_map = 1.0 / np.sqrt(weight_map)
        noise_map[noise_map == np.inf] = 1.0e8
        return NoiseMap(array=noise_map, pixel_scale=pixel_scale)


class PSF(ScaledSquarePixelArray):

    # noinspection PyUnusedLocal
    def __init__(self, array, pixel_scale, renormalize=False):
        """
        Class storing a 2D Point Spread Function (PSF), including its blurring kernel.

        Parameters
        ----------
        array : ndarray
            The 2d PSF blurring kernel.
        renormalize : bool
            Renormalize the PSF such that he sum of kernel values total 1.0?
        """

        # noinspection PyArgumentList
        super().__init__(array=array, pixel_scale=pixel_scale)
        if renormalize:
            self.renormalize()

    @classmethod
    def simulate_as_gaussian(cls, shape, pixel_scale, sigma, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0):
        """Simulate the PSF as an elliptical Gaussian profile."""
        from autolens.profiles.light_profiles import EllipticalGaussian
        gaussian = EllipticalGaussian(centre=centre, axis_ratio=axis_ratio, phi=phi, intensity=1.0, sigma=sigma)
        grid_1d = imaging_util.image_grid_1d_masked_from_mask_and_pixel_scales(mask=np.full(shape, False),
                                                                               pixel_scales=(pixel_scale, pixel_scale))
        gaussian_1d = gaussian.intensities_from_grid(grid=grid_1d)
        gaussian_2d = imaging_util.map_unmasked_1d_array_to_2d_array_from_array_1d_and_shape(array_1d=gaussian_1d,
                                                                                             shape=shape)
        return PSF(array=gaussian_2d, pixel_scale=pixel_scale, renormalize=True)

    @classmethod
    def from_fits_renormalized(cls, file_path, hdu, pixel_scale):
        """Loads a PSF from fits and renormalizes it

        Parameters
        ----------
        pixel_scale
        file_path: String
            The path to the file containing the PSF
        hdu : int
            The HDU the PSF is stored in the .fits file.

        Returns
        -------
        psf: PSF
            A renormalized PSF instance
        """
        psf = PSF.from_fits_with_scale(file_path, hdu, pixel_scale)
        psf.renormalize()
        return psf

    @classmethod
    def from_fits_with_scale(cls, file_path, hdu, pixel_scale):
        """
        Loads the PSF from a .fits file.

        Parameters
        ----------
        pixel_scale
        file_path: String
            The path to the file containing the PSF
        hdu : int
            The HDU the PSF is stored in the .fits file.
        """
        return cls(array=imaging_util.numpy_array_from_fits(file_path, hdu), pixel_scale=pixel_scale)

    def renormalize(self):
        """Renormalize the PSF such that its data_vector values sum to unity."""
        self[:, :] = np.divide(self, np.sum(self))

    def convolve(self, array):
        """
        Convolve an array with this PSF

        Parameters
        ----------
        array: ndarray
            An array representing the _data the PSF is convolved with.

        Returns
        -------
        convolved_array: ndarray
            An array representing the _data after convolution.

        Raises
        ------
        KernelException if either PSF psf dimension is odd
        """
        if self.shape[0] % 2 == 0 or self.shape[1] % 2 == 0:
            raise exc.KernelException("PSF Kernel must be odd")

        return scipy.signal.convolve2d(array, self, mode='same')


def setup_random_seed(seed):
    """Setup the random seed. If the input seed is -1, the code will use a random seed for every run. If it is \
    positive, that seed is used for all runs, thereby giving reproducible results.

    Parameters
    ----------
    seed : int
        The seed of the random number generator.
    """
    if seed == -1:
        seed = np.random.randint(0,
                                 int(1e9))  # Use one seed, so all regions have identical column non-uniformity.
    np.random.seed(seed)


def generate_poisson_noise(image, effective_exposure_map, seed=-1):
    """
    Generate a two-dimensional poisson noise_map-mappers from an _data.

    Values are computed from a Poisson distribution using the _data's input values in units of counts.

    Parameters
    ----------
    image : ndarray
        The 2D _data, whose values in counts are used to draw Poisson noise_map values.
    effective_exposure_map : Union(ndarray, int)
        2D array of the exposure time in each pixel used to convert to / from counts and electrons per second.
    seed : int
        The seed of the random number generator, used for the random noise_map maps.

    Returns
    -------
    poisson_noise_map: ndarray
        An array describing simulated poisson noise_map
    """
    setup_random_seed(seed)
    image_counts = np.multiply(image, effective_exposure_map)
    return image - np.divide(np.random.poisson(image_counts, image.shape), effective_exposure_map)


def load_imaging_from_fits(image_path, noise_map_path, psf_path, pixel_scale, image_hdu=0, noise_map_hdu=0, psf_hdu=0,
                           image_shape=None, psf_shape=None, noise_map_is_weight_map=False):
    data = ScaledSquarePixelArray.from_fits_with_pixel_scale(file_path=image_path, hdu=image_hdu,
                                                             pixel_scale=pixel_scale)

    noise_map = None

    if not noise_map_is_weight_map:
        noise_map = NoiseMap.from_fits_with_pixel_scale(file_path=noise_map_path, hdu=noise_map_hdu,
                                                        pixel_scale=pixel_scale)
    elif noise_map_is_weight_map:
        weight_map = Array.from_fits(file_path=noise_map_path, hdu=noise_map_hdu)
        noise_map = NoiseMap.from_weight_map(weight_map=weight_map, pixel_scale=pixel_scale)

    if image_shape is not None:
        data = data.resize_around_centre(new_shape=image_shape)
        noise_map = noise_map.resize_around_centre(new_shape=image_shape)

    psf = PSF.from_fits_with_scale(file_path=psf_path, hdu=psf_hdu, pixel_scale=pixel_scale)

    if psf_shape is not None:
        psf = psf.resize_around_centre(new_shape=psf_shape)

    return Image(array=data, pixel_scale=pixel_scale, psf=psf, noise_map=noise_map)


def load_imaging_from_path(image_path, noise_map_path, psf_path, pixel_scale, psf_trimmed_shape=None):
    data = ScaledSquarePixelArray.from_fits_with_pixel_scale(file_path=image_path, hdu=0, pixel_scale=pixel_scale)
    noise = ScaledSquarePixelArray.from_fits_with_pixel_scale(file_path=noise_map_path, hdu=0, pixel_scale=pixel_scale)
    psf = PSF.from_fits_with_scale(file_path=psf_path, hdu=0, pixel_scale=pixel_scale)
    if psf_trimmed_shape is not None:
        psf = psf.trim_around_centre(psf_trimmed_shape)
    return Image(array=data, pixel_scale=pixel_scale, psf=psf, noise_map=noise)


def output_imaging_to_fits(image, image_path, noise_map_path, psf_path, overwrite=False):
    imaging_util.numpy_array_to_fits(array=image, path=image_path, overwrite=overwrite)
    imaging_util.numpy_array_to_fits(array=image.noise_map, path=noise_map_path, overwrite=overwrite)
    imaging_util.numpy_array_to_fits(array=image.psf, path=psf_path, overwrite=overwrite)



from autolens.plotting import tools_array
from autolens.plotting import tools
from autolens.plotting import mapper_plotters

def plot_reconstructed_image(inversion, mask=None, positions=None, grid=None, as_subplot=False,
                             units='arcsec', kpc_per_arcsec=None, figsize=(7, 7), aspect='equal',
                             cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
                             cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                             title='Reconstructed Image', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
                             output_path=None, output_format='show', output_filename='reconstructed_image'):

    tools_array.plot_array(array=inversion.reconstructed_image, mask=mask, positions=positions, grid=grid,
                           as_subplot=as_subplot,
                           units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                           cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                           linthresh=linthresh, linscale=linscale,
                           cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                           title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                           xyticksize=xyticksize,
                           output_path=output_path, output_format=output_format, output_filename=output_filename)

def plot_reconstructed_pixelization(inversion, positions=None, should_plot_centres=False,
                                    should_plot_grid=False, image_pixels=None, source_pixels=None, as_subplot=False,
                                    units='arcsec', kpc_per_arcsec=None, figsize=(7, 7), aspect='equal',
                                    cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05,
                                    linscale=0.01,
                                    cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01,
                                    title='Reconstructed Pixelization', titlesize=16, xlabelsize=16, ylabelsize=16,
                                    xyticksize=16,
                                    output_path=None, output_format='show', output_filename='reconstructed_image'):

    tools.setup_figure(figsize=figsize, as_subplot=as_subplot)

    reconstructed_pixelization = \
        inversion.mapper.reconstructed_pixelization_from_solution_vector(inversion.solution_vector)

    tools_array.plot_array(array=reconstructed_pixelization, positions=positions, as_subplot=True,
                           units=units, kpc_per_arcsec=kpc_per_arcsec, figsize=figsize, aspect=aspect,
                           cmap=cmap, norm=norm, norm_min=norm_min, norm_max=norm_max,
                           linthresh=linthresh, linscale=linscale,
                           cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                           title=title, titlesize=titlesize, xlabelsize=xlabelsize, ylabelsize=ylabelsize,
                           xyticksize=xyticksize,
                           output_filename=output_filename)

    mapper_plotters.plot_rectangular_mapper(mapper=inversion.mapper, should_plot_centres=should_plot_centres,
                                            should_plot_grid=should_plot_grid,
                                            image_pixels=image_pixels, source_pixels=source_pixels,
                                            as_subplot=True,
                                            units=units, kpc_per_arcsec=kpc_per_arcsec,
                                            title=title, titlesize=titlesize, xlabelsize=xlabelsize,
                                            ylabelsize=ylabelsize, xyticksize=xyticksize)

    tools.output_figure(array=reconstructed_pixelization, as_subplot=as_subplot, output_path=output_path,
                        output_filename=output_filename, output_format=output_format)
    tools.close_figure(as_subplot=as_subplot)



from profiling import tools

from autolens import exc


class Pixelization(object):

    def __init__(self, pixels, regularization_coefficients=(1.0,), pix_signal_scale=1.0):
        """
        Abstract base class for a inversion, which discretizes a set of masked_image and sub grid grid into \
        pixels. These pixels then incorrect_fit a  data_vector-set using a linear inversion, where their regularization_matrix matrix
        enforces smoothness between pixel values.

        A number of 1D and 2D arrays are used to represent mappings betwen masked_image, sub, pix, and cluster pixels. The \
        nomenclature here follows grid_to_grid, such that it maps the index of a value on one grid to another. For \
        howtolens:

        - pix_to_image[2] = 5 tells us that the 3rd inversion-pixel maps to the 6th masked_image-pixel.
        - sub_to_pixelization[4,2] = 2 tells us that the 5th sub-pixel maps to the 3rd inversion-pixel.

        Parameters
        ----------
        pixels : int
            The number of pixels in the inversion.
        regularization_coefficients : (float,)
            The regularization_matrix coefficients used to smooth the pix reconstructed_image.
        pix_signal_scale : float
            A hyper-parameter which scales the signal attributed to each pixel, used for weighted regularization_matrix.
        """
        self.pixels = pixels
        self.regularization_coefficients = regularization_coefficients
        self.pix_signal_scale = pix_signal_scale


class Rectangular(Pixelization):

    def __init__(self, shape=(3, 3), regularization_coefficients=(1.0,)):
        """A rectangular inversion where pixels appear on a Cartesian, uniform and rectangular grid \
        of  shape (rows, columns).

        Like an masked_image grid, the indexing of the rectangular grid begins in the top-left corner and goes right and down.

        Parameters
        -----------
        shape : (int, int)
            The dimensions of the rectangular grid of pixels (x_pixels, y_pixel)
        regularization_coefficients : (float,)
            The regularization_matrix coefficients used to smooth the pix reconstructed_image.
        """

        if shape[0] <= 2 or shape[1] <= 2:
            raise exc.PixelizationException('The rectangular inversion must be at least dimensions 3x3')

        super(Rectangular, self).__init__(shape[0] * shape[1], regularization_coefficients)

        self.shape = shape

    def neighbors_from_pixelization(self):
        """Compute the neighbors of every pixel as a list of the pixel index's each pixel shares a vertex with.

        The uniformity of the rectangular grid's geometry is used to compute this.
        """

        def compute_corner_neighbors(pix_neighbors):

            pix_neighbors[0] = [1, self.shape[1]]
            pix_neighbors[self.shape[1] - 1] = [self.shape[1] - 2, self.shape[1] + self.shape[1] - 1]
            pix_neighbors[self.pixels - self.shape[1]] = [self.pixels - self.shape[1] * 2,
                                                          self.pixels - self.shape[1] + 1]
            pix_neighbors[self.pixels - 1] = [self.pixels - self.shape[1] - 1, self.pixels - 2]

            return pix_neighbors

        def compute_top_edge_neighbors(pix_neighbors):

            for pix in range(1, self.shape[1] - 1):
                pixel_index = pix
                pix_neighbors[pixel_index] = [pixel_index - 1, pixel_index + 1, pixel_index + self.shape[1]]

            return pix_neighbors

        def compute_left_edge_neighbors(pix_neighbors):

            for pix in range(1, self.shape[0] - 1):
                pixel_index = pix * self.shape[1]
                pix_neighbors[pixel_index] = [pixel_index - self.shape[1], pixel_index + 1, pixel_index + self.shape[1]]

            return pix_neighbors

        def compute_right_edge_neighbors(pix_neighbors):

            for pix in range(1, self.shape[0] - 1):
                pixel_index = pix * self.shape[1] + self.shape[1] - 1
                pix_neighbors[pixel_index] = [pixel_index - self.shape[1], pixel_index - 1, pixel_index + self.shape[1]]

            return pix_neighbors

        def compute_bottom_edge_neighbors(pix_neighbors):

            for pix in range(1, self.shape[1] - 1):
                pixel_index = self.pixels - pix - 1
                pix_neighbors[pixel_index] = [pixel_index - self.shape[1], pixel_index - 1, pixel_index + 1]

            return pix_neighbors

        def compute_central_neighbors(pix_neighbors):

            for x in range(1, self.shape[0] - 1):
                for y in range(1, self.shape[1] - 1):
                    pixel_index = x * self.shape[1] + y
                    pix_neighbors[pixel_index] = [pixel_index - self.shape[1], pixel_index - 1, pixel_index + 1,
                                                  pixel_index + self.shape[1]]

            return pix_neighbors

        pixel_neighbors = [[] for _ in range(self.pixels)]

        pixel_neighbors = compute_corner_neighbors(pixel_neighbors)
        pixel_neighbors = compute_top_edge_neighbors(pixel_neighbors)
        pixel_neighbors = compute_left_edge_neighbors(pixel_neighbors)
        pixel_neighbors = compute_right_edge_neighbors(pixel_neighbors)
        pixel_neighbors = compute_bottom_edge_neighbors(pixel_neighbors)
        pixel_neighbors = compute_central_neighbors(pixel_neighbors)

        return pixel_neighbors


@tools.tick_toc_x10
def solution_10x10():
    pix = Rectangular(shape=(10, 10))
    pix.neighbors_from_pixelization()


@tools.tick_toc_x10
def solution_20x20():
    pix = Rectangular(shape=(20, 20))
    pix.neighbors_from_pixelization()


@tools.tick_toc_x10
def solution_30x30():
    pix = Rectangular(shape=(30, 30))
    pix.neighbors_from_pixelization()


@tools.tick_toc_x10
def solution_40x40():
    pix = Rectangular(shape=(40, 40))
    pix.neighbors_from_pixelization()


@tools.tick_toc_x10
def solution_50x50():
    pix = Rectangular(shape=(50, 50))
    pix.neighbors_from_pixelization()


@tools.tick_toc_x10
def solution_100x100():
    pix = Rectangular(shape=(100, 100))
    pix.neighbors_from_pixelization()


if __name__ == "__main__":
    solution_10x10()
    solution_20x20()
    solution_30x30()
    solution_40x40()
    solution_50x50()
    solution_100x100()
